# Copyright 2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
:mod:`task` stores the :class:`~giza.task.Task()` and
:class:`~giza.task.MapTask()` classes which represent single units of work in
the context of a :class:`~giza.app.BuildApp()` procedure.

:mod:`task` also has a number of dependency resolution helper functions.
"""

import logging
import sys
import os.path
import collections

from giza.libgiza.config import ConfigurationBase

logger = logging.getLogger('giza.libgiza.task')

if sys.version_info >= (3, 0):
    basestring = str


class Task(object):
    """
    Provides a common interface for defining an operational unit of work in a
    concurrent :class:`~giza.app.BuildApp()` environment.

    With :attr:`~giza.task.Task.target` and :attr:`~giza.task.Task.dependency`
    defined, if a ``target`` file exists and was modified after the
    ``dependency`` file, the :class:`~giza.task.Task()` operation becomes a
    no-op, unless forced.
    """

    def __init__(self, job=None, args=None,
                 description=None, target=None, dependency=None, ignore=None):
        """
        All arguments are optional. You can define a :class:`~giza.task.Task()`
        either upon creation, or after creation by modifying attributes.

        :param callable job: A callable object that the task will execute.

        :param args: An argument list or mapping.

        :param string description: Describes the task. Used in error messages.

        :param string target: A file name. A path to a file that the task will create.

        :param string dependency: A file name. A path to a file that the task
           depends on. When specified, the task will only run if forced or if
           the ``depdendency`` file is newer than the target file.
        """

        self.spec = {}
        self._conf = None
        self._args = None
        self._force = None
        self._ignore_errors = None
        self._description = None
        if job is not None:
            self.job = job
        self._finalizers = []
        self.args_type = None

        self.target = target
        self.dependency = dependency

        # use default setter logic for these options
        if args is not None:
            self.args = args
        if ignore is not None:
            self.ignore_errors = ignore
        if description is not None:
            self.description = description

        logger.debug('created task object calling {0}, for {1}'.format(job, description))
        self._task_id = None

    @property
    def task_id(self):
        if self._task_id is None:
            self._task_id = hash(str(self.job)) + hash(str(self.args))

        return self._task_id

    @property
    def description(self):
        if self._description is None:
            return str(self.job)
        else:
            return ' '.join([self._description, str(self.job)])

    @description.setter
    def description(self, value):
        self._description = value

    @property
    def dependency(self):
        return self._dependency

    @dependency.setter
    def dependency(self, value):
        self._dependency = value

    @property
    def target(self):
        return self._target

    @target.setter
    def target(self, value):
        self._target = value

    @property
    def force(self):
        if self._force is None:
            if self.conf is None:
                return False
            else:
                logger.warning('deprecated use of conf object in app setup for force value')
                return self.conf.runstate.force

        return self._force

    @force.setter
    def force(self, value):
        if isinstance(value, bool):
            self._force = value

    @property
    def ignore_errors(self):
        if self._ignore_errors is None:
            if self.conf is None:
                return True
            else:
                logger.warning('deprecated use of conf object in app setup for ignore_errors value')
                return self.conf.runstate.ignore_errors

        return False

    @ignore_errors.setter
    def ignore_errors(self, value):
        if isinstance(value, bool):
            self._ignore_errors = value

    def define_dependency_node(self, target, dependency):
        self.target = target
        self.dependency = dependency

    @property
    def conf(self):
        return self._conf

    @conf.setter
    def conf(self, value):
        if value is None:
            pass
        elif isinstance(value, ConfigurationBase):
            self._conf = value
        else:
            raise TypeError

    @property
    def job(self):
        return self.spec['job']

    @job.setter
    def job(self, value):
        if isinstance(value, collections.Callable):
            self.spec['job'] = value
        else:
            raise TypeError

    @property
    def args(self):
        return self._args

    @args.setter
    def args(self, value):
        if isinstance(value, dict):
            self.args_type = 'kwargs'
            self._args = value
        elif isinstance(value, (list, tuple)):
            self.args_type = 'args'
            self._args = value
        elif isinstance(value, basestring):
            self.args_type = 'args'
            self._args = [value]
        else:
            logger.critical(type(value))

    @property
    def finalizers(self):
        return self._finalizers

    @finalizers.setter
    def finalizers(self, value):
        if (isinstance(value, tuple) and
                len(value) == 2 and
                isinstance(value[1], Task)):
            self._finalizers.append(value)
        elif hasattr(value, 'queue'):
            raise TypeError(type(value), len(value.queue))
        elif hasattr(value, 'run'):
            self._finalizers.append(value)
        elif isinstance(value, collections.Iterable):
            for task in value:
                if hasattr(task, 'queue'):
                    raise TypeError(type(task), len(task.queue))
                elif hasattr(task, 'run'):
                    self._finalizers.append(task)
                elif isinstance(task, tuple) and isinstance(task[1], Task):
                    self._finalizers.append(task)
                elif isinstance(task, collections.Iterable):
                    raise TypeError(type(task), task)
        else:
            raise TypeError(type(value))

    def add_finalizer(self, task):
        self.finalizers = task

        return task

    @property
    def needs_rebuild(self):
        """
        Used by the execution application to see if a rebuild is needed. Always
        returns ``True`` if there is no target or when running in *force* mode,
        otherwise checks the ``mtime`` of the files using
        :func:`giza.libgiza.task.check_dependency()`.
        """

        if self.target is None:
            return True
        elif self.dependency is None:
            return True
        elif self.force is True:
            return True
        else:
            return check_dependency(self.target, self.dependency)

    def run(self):
        logger.debug('({0}) calling {1}'.format(self.task_id, self.job))
        if self.args_type == 'kwargs':
            result = self.job(**self.args)
        elif self.args_type == 'args':
            result = self.job(*self.args)
        else:
            result = self.job()

        logger.debug('completed running task {0}, {1}'.format(self.task_id, self.description))

        return result

    def finalize(self):
        logger.debug('running {0} finalizers (serially)'.format(len(self.finalizers)))
        results = []

        for task in self.finalizers:
            results.append(task.run())

            if len(task.finalizers) > 0:
                results.extend(task.finalize())

        logger.debug('completed {0} finalizers'.format(len(results)))
        return results


class MapTask(Task):
    """
    A variant of :class:`~giza.task.Task()` that defines a task that like the
    kind of operation that would run in a :func:`map()` function, processing the
    contents of an operable with a single function.
    """

    def __init__(self, job=None, description=None, target=None, dependency=None):
        super(MapTask, self).__init__(job=job, description=description,
                                      target=target, dependency=dependency)
        self._iter = []

    @property
    def iter(self):
        return self._iter

    @iter.setter
    def iter(self, value):
        if isinstance(value, collections.Iterable):
            self._iter = value
        else:
            raise TypeError

    def run(self):
        return list(map(self.job, self.iter))

# Dependency Checking


def check_dependency(target, dependency):
    """
    Determines if a target requires rebuilding based on a provided
    dependency.

    :param string target: A file name or list of file names.
    :type target: string, list

    :param dependency: A file name or list of file names.
    :type dependency: string, list

    :returns: A boolean. If either the ``target`` or ``dependency`` doesn't
       exist, or if the ``target`` was modified more recently than the
       ``dependency`` returns ``True``; otherwise ``False``.

    :func:`~giza.task.check_dependency()` Accepts dependencies in the form of a
    single file name, or as a list, and will return ``True`` if *any* dependent
    file is newer than the target. Also returns ``True`` when:

    - ``target`` or ``dependency`` is ``None``.

    - ``target`` or ``dependency`` does not exist.
    """

    if dependency is None:
        return True
    elif target is None:
        return True
    elif isinstance(target, list):
        for t in target:
            if os.path.exists(t) is False:
                return True
            else:
                return check_dependency(t, dependency)
    elif os.path.exists(target) is False:
        return True
    elif isinstance(dependency, list):
        target_time = os.path.getmtime(target)
        for dep in dependency:
            if dep is None:
                return True
            elif target_time < os.path.getmtime(dep):
                logger.debug("rebuild of {0} triggered by: {1}".format(target, dep))
                return True
        return False
    elif os.path.exists(dependency):
        if os.path.getmtime(target) < os.path.getmtime(dependency):
            logger.debug("rebuild of {0} triggered by: {1}".format(target, dependency))
            return True
        else:
            return False
    else:
        logger.error('{0} is not a valid dependency'.format(dependency))
        return True



# Copyright 2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os.path

from giza.inheritance import DataContentBase, DataCache
from giza.content.steps.models import StepData

logger = logging.getLogger('giza.content.steps.inheritance')


class StepFile(DataContentBase):
    content_class = StepData

    def add(self, doc):
        content = super(StepFile, self).add(doc)

        obj = self.content[content.ref]

        if 'number' not in obj:
            if len(self.content) == 1:
                obj.number = 1
            else:
                obj.number = self.fetch(self.ordering[-1]).number + 1

        return obj

    def target(self, fn):
        # fn is the source file
        return os.path.join(self.conf.system.content.steps.output_dir,
                            self.conf.system.content.steps.get_basename(fn)) + '.rst'


class StepDataCache(DataCache):
    content_class = StepFile
    content_type = 'steps'



import requests
import logging

from requests.auth import HTTPDigestAuth

from giza.config.credentials import CredentialsConfig

logger = logging.getLogger('giza.corp')


def corp_api_call(endpoint, conf):
    credentials = CredentialsConfig(conf.site.credentials).corp

    url = conf.site.corp + '/api/' + endpoint
    headers = {'accept': 'application/json'}
    response = requests.get(url,
                            auth=HTTPDigestAuth(credentials.username, credentials.password),
                            headers=headers)

    return response.json()


def get_contributor_list(conf):
    contributors = corp_api_call('contributors', conf)['contributors']

    c_github = list({str(c['github_username']) for c in contributors
                     if ('github_username' in c and
                         c['github_username'] is not None and
                         c['github_username'] != u'')})

    return c_github



# Copyright 2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Contains schema, type checking, and validation structures for the data in
example files.
"""

import logging

from giza.libgiza.config import ConfigurationBase
from giza.libgiza.inheritance import InheritableContentError

from giza.inheritance import InheritableContentBase
from giza.content.helper import get_all_languages

logger = logging.getLogger('giza.content.examples.models')


class ExampleData(InheritableContentBase):

    @property
    def collection(self):
        if 'collection' not in self.state:
            return None
        else:
            return self.state['collection']

    @collection.setter
    def collection(self, value):
        self.state['collection'] = value

    @property
    def name(self):
        return self.collection

    @property
    def ref(self):
        return self.collection

    @property
    def documents(self):
        return self.state['documents']

    @documents.setter
    def documents(self, value):
        if isinstance(value, list):
            self.state['documents'] = value
        else:
            raise TypeError('{0} is not a list'.format(value))

    @property
    def options(self):
        if 'options' not in self.state:
            return ExampleOptions()
        else:
            return self.state['options']

    @options.setter
    def options(self, value):
        self.state['options'] = ExampleOptions(value)


class ExampleOptions(ConfigurationBase):

    @property
    def show_title(self):
        if 'title' not in self.state:
            return True
        else:
            return self.state['title']

    @show_title.setter
    def show_title(self, value):
        if isinstance(value, bool):
            self.state['title'] = value
        else:
            raise TypeError

    @property
    def show_collection(self):
        if 'collection' not in self.state:
            return True
        else:
            return self.state['collection']

    @show_collection.setter
    def show_collection(self, value):
        if isinstance(value, bool):
            self.state['collection'] = value
        else:
            raise TypeError

    @property
    def base_file(self):
        if 'is_base_file' not in self.state:
            return False
        else:
            return self.state['is_base_file']

    @base_file.setter
    def base_file(self, value):
        if isinstance(value, bool):
            self.state['is_base_file'] = value
        else:
            raise TypeError


class ExampleOperationBlock(ConfigurationBase):
    _option_registry = ['pre', 'post', 'final']

    @property
    def code(self):
        return self.state['code']

    @code.setter
    def code(self, value):
        if isinstance(value, list):
            self.state['code'] = value
        else:
            self.state['code'] = value.split('\n')

    @property
    def content(self):
        return self.state['content']

    @content.setter
    def content(self, value):
        self.state['content'] = value.split('\n')

    @property
    def literalinclude(self):
        return self.state['literalinclude']

    @literalinclude.setter
    def literalinclude(self, value):
        self.state['literalinclude'] = value

    @property
    def language(self):
        return self.state['language']

    @language.setter
    def language(self, value):
        if value in get_all_languages():
            self.state['language'] = value
        else:
            m = '{0} is not a supported language'.format(value)
            logger.error(m)
            TypeError(m)


class ExampleCase(InheritableContentBase):

    @property
    def operation(self):
        if 'operation' in self.state:
            return self.state['operation']
        elif self.ref.startswith("_"):
            return ""
        else:
            raise InheritableContentError('no operation specified in: ' + self.ref)

    @operation.setter
    def operation(self, value):
        if isinstance(value, list):
            self.state['operation'] = [ExampleOperationBlock(doc)
                                       for doc in value]
        elif isinstance(value, dict):
            self.state['operation'] = [ExampleOperationBlock(value)]
        else:
            m = 'unable to create operation block from {0}'.format(value)
            logger.error(m)
            TypeError(m)

    name = operation

    @property
    def results(self):
        return self.state['results']

    @results.setter
    def results(self, value):
        if value is None:
            self.state['results'] = value
        elif isinstance(value, list):
            v = []
            for ln in value:
                v.extend(ln.split('\n'))

            self.state['results'] = v
        else:
            self.state['results'] = value.split('\n')



# Copyright 2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Provides the :class:`~giza.git.GitRepo()` class that provides a thin
Python-layer on top of common git operations.
"""

import logging
import os
import re
import sys
import contextlib
import subprocess
import shlex

logger = logging.getLogger('giza.libgiza.git')

if sys.version_info >= (3, 0):
    basestring = str

    def unicode(s):
        return str(s, 'utf-8')


class GitError(Exception):
    pass


class GitRepo(object):
    """
    An object that represents a git repository, and provides methods to wrap
    many common git commands and basic aggregate operations.
    """

    def __init__(self, path=None):
        """
        :param string path: Optional. Defines a the path of the git
           repository. If not specified, defaults to the current working
           directory.

           Attempts to find the top-level (root) directory of the git
           repository, if you specify a path that's a sub-directory in a git
           repository.
        """

        if path is None:
            self.path = os.getcwd()
        else:
            self.path = path

        logger.debug("created git repository management object for {0}".format(self.path))

    def cmd(self, *args):
        cmd_parts = ['git']

        for arg in args:
            if isinstance(arg, list):
                cmd_parts.extend(arg)
            else:
                cmd_parts.append(arg)

        try:
            if os.path.exists(self.path):
                logger.debug("running git command ({0}) at path {1}".format(' '.join(cmd_parts),
                                                                            self.path))
                return unicode(subprocess.check_output(args=cmd_parts,
                                                       cwd=self.path,
                                                       stderr=subprocess.STDOUT).strip())
            else:
                logger.debug("running git command: " + ' '.join(cmd_parts))
                return unicode(subprocess.check_output(args=cmd_parts,
                                                       stderr=subprocess.STDOUT).strip())

        except Exception as e:
            raise GitError('encountered error {0} ({1}) with {2} in repository '
                           '{3}'.format(e, type(e), ' '.join(cmd_parts), self.path))

    def clone(self, remote, repo_path=None, branch=None, depth=None):
        args = ['clone', remote]

        if branch is not None:
            args.extend(['--branch', branch])

        if depth is not None:
            if isinstance(depth, int):
                depth = str(int)
            args.extend(["--depth", depth])

        if repo_path is not None:
            args.append(repo_path)
            self.path = repo_path

        ret = self.cmd(*args)

        return ret

    def create_repo(self, bare=False):
        args = ['init']

        if bare is True:
            args.append('--bare')

        return self.cmd(*args)

    def top_level(self):
        try:
            return self.cmd('rev-parse', '--show-toplevel')
        except GitError:
            logger.error('{0} may not be a git repository'.format(self.path))
            return self.path

    def is_repository(self):
        if os.path.isdir(os.path.join(self.top_level(), ".git")):
            return True
        else:
            return False

    def remotes(self):
        return self.cmd('remote').split('\n')

    def author_email(self, sha=None):
        if sha is None:
            sha = self.sha()

        return self.cmd('log', shlex.split(sha + '~..' + sha + " --pretty='format:%ae'"))

    def branch_exists(self, name):
        r = self.cmd('branch', '--list', name).split('\n')
        if '' in r:
            r.remove('')

        if name in r:
            return True
        else:
            return False

    def branch_file(self, path, branch='master'):
        return self.cmd('show', ':'.join((branch, path)))

    def checkout(self, ref):
        self.cmd('checkout', ref)
        return True

    def create_branch(self, name, tracking=None):
        args = ['branch', name]

        if tracking is not None:
            args.append(tracking)

        return self.cmd(*args)

    def checkout_branch(self, name, tracking=None):
        if self.current_branch() == name:
            return

        args = ['checkout']

        if not self.branch_exists(name):
            args.append('-b')

        args.append(name)

        if tracking is not None:
            args.append(tracking)

        return self.cmd(*args)

    def remove_branch(self, name, force=False):
        args = ['branch']

        if force is False:
            args.append('-d')
        else:
            args.append('-D')

        args.append(name)

        return self.cmd(*args)

    def rebase(self, onto):
        return self.cmd('rebase', onto)

    def merge(self, branch):
        return self.cmd('merge', branch)

    def hard_reset(self, ref='HEAD'):
        return self.cmd('reset', '--hard', ref)

    def reset(self, ref='HEAD'):
        return self.cmd('reset', ref)

    def fetch(self, remote='origin'):
        return self.cmd('fetch', remote)

    def fetch_all(self):
        return self.cmd('fetch', '--all')

    def update(self):
        return self.cmd('pull', '--rebase')

    def pull(self, remote='origin', branch='master'):
        return self.cmd('pull', '--ff-only', remote, branch)

    def push(self, remote='origin', ref=None):
        if ref is None:
            ref = self.current_branch()

        return self.cmd("push", remote, ref)

    def tag(self, name, ref="HEAD", annotation=None, delete=False, force=False):
        args = ["tag"]

        if delete is True:
            args.append("-d")

        if force is True:
            args.append("-f")

        if annotation is not None:
            if isinstance(annotation, basestring):
                args.extend(["-m", annotation])
            else:
                raise TypeError("tag annotations must be strings. {0} {1} "
                                "is not a string".format(annotation, type(annotation)))

        args.extend([name, ref])

        return self.cmd(*args)

    def is_tagged(self, name, ref="HEAD", lightweight=False):
        cmd = ["describe"]

        if lightweight is True:
            cmd.append("--tags")

        cmd.append(name)

        try:
            self.cmd(*cmd)
        except GitError:
            return False

        if ref == "HEAD":
            ref = self.sha(ref)

        tagged_commit = self.cmd(["rev-list", "-n", "1", name])
        if tagged_commit == ref:
            return True
        else:
            return False

    def current_branch(self):
        return self.cmd('rev-parse', '--abbrev-ref', 'HEAD')

    def sha(self, ref='HEAD'):
        return self.cmd('rev-parse', '--verify', ref)

    def commit_messages(self, num=1):
        args = ['log', '--oneline', '--max-count=' + str(num)]
        log = self.cmd(*args)

        return [' '.join(m.split(' ')[1:])
                for m in log.split('\n')]

    def cherry_pick(self, *args):
        if len(args) == 1:
            args = args[0]

        for commit in args:
            self.cmd('cherry-pick', commit)
            logger.info('cherry picked ' + commit)

    def am(self, patches, repo=None, sign=False):
        cmd_base = 'curl -s {path} | git am --3way'

        if sign is True:
            cmd_base += ' --signoff'

        for obj in patches:
            if obj.startswith('http'):
                path = obj
                if not obj.endswith('.patch'):
                    path = obj + '.patch'

                logger.info("applying {0}".format(path))
            elif re.search('[a-zA-Z]+', obj):
                path = '/'.join([repo, 'commit', obj]) + '.patch'

                logger.info('merging {0} for {1} into {2}'.format(obj, repo, self.current_branch()))
            else:
                if repo is None:
                    logger.warning('not applying "{0}", because of missing repo'.format(obj))
                    continue
                else:
                    path = '/'.join([repo, 'pull', obj]) + '.patch'
                    logger.info("applying {0}".format(path))

            logger.info(cmd_base.format(path=path))
            subprocess.call(cwd=self.path,
                            args=cmd_base.format(path=path),
                            stderr=subprocess.STDOUT,
                            shell=True)

    @contextlib.contextmanager
    def branch(self, name):
        starting_branch = self.current_branch()

        if name != starting_branch:
            self.checkout(name)

        yield

        if name != starting_branch:
            self.checkout(starting_branch)



# Copyright 2015 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
The complete goal for this component is to take the results of a Jira query and
generate RST with the contents of that change log at release so that we can
remove several manual steps from the larger release project.

Requirements/Features:

1. For every minor release, the system will create a single file that the
   publication/push manager will need to review, edit, and commit.

   - The /release-notes/<release-series>-changelog.txt file can be generated
     from template if it doesn't exist.

   - There will be an intermediate generated file that we will be ignored from
     git regenerated on as needed? and will include all of the generated files
     with each version so that.

   - The system will do auto-backfill for all of the covered versions *if* the
     files don't exist.

2. Configuration for the properties (i.e. components, projects, JIRA instances,)
   should be in the global (ie. docs-tools/data) directory. Configuration for
   actual generation (file locations, paths, etc.) should pull from the
   site configuration object and other config should be a linked configuration
   file.

3. Builds of the documentation should *not* (re)generate change-logs for new
   entries, nor should they expect that JIRA be available or responsive.

Implementation Steps:

- Model new configuration data.
- Generate new pages:
  - change log master page
  - intermediate include file
  - minor release change log.
- Create severable tasks.
- Add entry points.
- Convert existing data on the 2.6 and 3.0 (master) branches of the documentation.
"""



# Copyright 2015 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os.path

from giza.libgiza.task import Task

from giza.content.post.json_output import json_output_tasks
from giza.content.post.singlehtml import finalize_single_html_tasks
from giza.content.post.archives import man_tarball, html_tarball, get_tarball_name
from giza.content.post.manpages import manpage_url_tasks
from giza.content.post.gettext import gettext_tasks
from giza.content.post.slides import slide_tasks
from giza.content.post.latex import pdf_tasks
from giza.content.post.sites import (finalize_epub_build,
                                     finalize_dirhtml_build, error_pages)

logger = logging.getLogger('giza.content.post.sphinx')


def printer(string):
    logger.info(string)


def finalize_sphinx_build(sconf, conf):
    target = sconf.builder

    tasks = []
    if target == 'html' and not conf.runstate.fast:
        t = Task(job=html_tarball,
                 args=(sconf.name, sconf.build_output, conf),
                 target=[get_tarball_name('html', conf),
                         get_tarball_name('link-html', conf)],
                 dependency=None,
                 description="creating tarball for html archive")
        tasks.append(t)
    elif target == 'dirhtml' and not conf.runstate.fast:
        # We're experiencing some cases were giza seemingly randomly doesn't migrate.
        # Log this to help us figure out what's going on.
        logger.info('Going to migrate {} to {}'.format(
            sconf.fq_build_output,
            os.path.join(conf.paths.projectroot, conf.paths.public_site_output)))

        for job in (finalize_dirhtml_build, error_pages):
            t = Task(job=job,
                     args=(sconf, conf),
                     target=os.path.join(conf.paths.projectroot, conf.paths.public_site_output),
                     dependency=None)
            tasks.append(t)
    elif target == 'epub':
        t = Task(job=finalize_epub_build,
                 args=(target, conf),
                 description='finalizing epub build',
                 dependency=None,
                 target=True)
        tasks.append(t)
    elif target == 'man':
        t = Task(job=man_tarball,
                 args=(sconf.name, sconf.build_output, conf),
                 target=[get_tarball_name('man', conf),
                         get_tarball_name('link-man', conf)],
                 dependency=None,
                 description="creating tarball for manpages")

        tasks.extend(manpage_url_tasks(target, conf))
        tasks.append(('final', t))
    elif target == 'slides' and not conf.runstate.fast:
        tasks.extend(slide_tasks(sconf, conf))
    elif target == 'json':
        json_tasks, transfer_op = json_output_tasks(conf)
        tasks.extend(json_tasks)
        tasks.append(('final', transfer_op))  # this is less than ideal
    elif target == 'singlehtml':
        tasks.extend(finalize_single_html_tasks(target, conf))
    elif target == 'latex':
        tasks.extend(pdf_tasks(sconf, conf))
    elif target == 'gettext':
        tasks.extend(gettext_tasks(conf))
    elif target == 'linkcheck':
        msg_str = '{0}: See {1}/{0}/output.txt for output.'
        t = Task(job=printer,
                 args=[msg_str.format(target, conf.paths.branch_output)],
                 target=os.path.join(conf.paths.projectroot,
                                     conf.paths.branch_output, target, 'output.txt'),
                 dependency=None)
        tasks.append(t)

    logger.debug('adding {0} finalizing tasks for {1} build'.format(len(tasks), target))
    return tasks



# Copyright 2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numbers

from unittest import TestCase

from giza.libgiza.task import MapTask, Task
from giza.libgiza.app import BuildApp
from giza.config.main import Configuration
from giza.config.runtime import RuntimeStateConfig


class BaseTaskSuite(object):
    def test_configuration_object_validation_rejection(self):
        for i in [1, 'config', {'config': 1}]:
            t = self.Task()

            with self.assertRaises(TypeError):
                t.conf = i

    def test_configuration_object_validation_acceptance(self):
        for i in [self.c, Configuration(), RuntimeStateConfig()]:
            t = self.Task()

            self.assertIsNone(t.conf)
            t.conf = i
            self.assertIs(i, t.conf)

    def test_job_validation_rejection(self):
        for i in [None, True, False, 1, 'foo']:
            t = self.Task()

            with self.assertRaises(TypeError):
                t.job = i

    def test_job_validation_acceptance(self):
        for i in [Configuration, sum, TypeError, map]:
            t = self.Task()
            t.job = i

            self.assertIs(i, t.job)

    def test_args_type_setting_dict(self):
        self.assertIsNone(self.task._args)
        self.assertIsNone(self.task.args_type)

        self.task.args = {'foo': 1}
        self.assertEqual(self.task.args_type, 'kwargs')

    def test_args_type_setting_args(self):
        self.assertIsNone(self.task._args)
        self.assertIsNone(self.task.args_type)

        self.task.args = [1, 2, 3, 4]
        self.assertEqual(self.task.args_type, 'args')

    def test_args_type_setting_tuple(self):
        self.assertIsNone(self.task._args)
        self.assertIsNone(self.task.args_type)

        self.task.args = (1, 2, 3, 4)
        self.assertEqual(self.task.args_type, 'args')

    def test_args_type_setting_string(self):
        self.assertIsNone(self.task._args)
        self.assertIsNone(self.task.args_type)

        self.task.args = 'None'
        self.assertEqual(self.task.args_type, 'args')

    def test_rebuild_requirment_checker_base(self):
        self.assertIsNone(self.task.target)
        self.assertIsNone(self.task.dependency)
        self.assertFalse(self.c.runstate.force)

        self.assertTrue(self.task.needs_rebuild)

    def test_rebuild_requirment_checker_dep(self):
        self.task.target = True

        self.assertIsNotNone(self.task.target)
        self.assertIsNone(self.task.dependency)
        self.assertFalse(self.c.runstate.force)

        self.assertTrue(self.task.needs_rebuild)

    def test_rebuild_requirment_checker_force(self):
        self.task.target = True
        self.task.dependency = True
        self.c.runstate.force = True

        self.assertIsNotNone(self.task.target)
        self.assertIsNotNone(self.task.dependency)
        self.assertTrue(self.c.runstate.force)

        self.assertTrue(self.task.needs_rebuild)

    def test_finalizer_setter_error_app(self):
        with self.assertRaises(TypeError):
            self.task.finalizers = BuildApp()

    def test_finalizer_setter_error_app_in_list(self):
        with self.assertRaises(TypeError):
            self.task.finalizers = [self.Task(), BuildApp()]

    def test_add_finalizer_returns_task(self):
        self.assertTrue(len(self.task.finalizers) == 0)
        t = self.Task()
        check = self.task.add_finalizer(t)
        self.assertIs(t, check)

    def test_finalizer_setter_error_fallthrough(self):
        with self.assertRaises(TypeError):
            self.task.finalizers = 1

    def test_finalizer_setter_error_nested_list(self):
        with self.assertRaises(TypeError):
            self.task.finalizers = [self.Task(), [self.Task(), self.Task()]]

    def test_finalizer_setter_non_distructive_assignment(self):
        self.assertEqual(len(self.task.finalizers), 0)
        self.task.finalizers = self.Task()
        self.assertEqual(len(self.task.finalizers), 1)
        self.task.finalizers = self.Task()
        self.assertEqual(len(self.task.finalizers), 2)
        self.task.finalizers = [self.Task(), self.Task(), self.Task(), self.Task()]
        self.assertEqual(len(self.task.finalizers), 6)

    def test_finalizer_setter_special_case_for_tuples_error(self):
        self.assertEqual(len(self.task.finalizers), 0)
        with self.assertRaises(TypeError):
            self.task.finalizers = (self.Task(), 'keyword')

    def test_finalizer_setter_special_case_for_tuples_list_error(self):
        self.assertEqual(len(self.task.finalizers), 0)
        with self.assertRaises(TypeError):
            self.task.finalizers = [(self.Task(), 'keyword'), self.Task()]

    def test_finalizer_setter_special_case_for_tuples(self):
        self.assertEqual(len(self.task.finalizers), 0)
        self.task.finalizers = ('keyword', self.Task())
        self.assertEqual(len(self.task.finalizers), 1)

    def test_finalizer_setter_special_case_for_tuples_list(self):
        self.assertEqual(len(self.task.finalizers), 0)
        self.task.finalizers = [('keyword', self.Task()), self.Task()]
        self.assertEqual(len(self.task.finalizers), 2)


class TestTask(BaseTaskSuite, TestCase):
    @classmethod
    def setUp(self):
        self.c = Configuration()
        self.c.runstate = RuntimeStateConfig()
        self.task = Task()
        self.task.job = sum
        self.Task = Task
        self.task.conf = self.c

    def test_running_tasks(self):
        def sum_func(iterable, start):
            return sum(iterable, start)

        for i in (((1, 2, 3), 0), ([1, 2, 3], 0), {'iterable': [1, 2, 3], 'start': 0}):
            t = self.Task()

            t.job = sum_func
            t.args = i
            t.description = "test task: " + str(i)

            self.assertEqual(t.run(), 6)

    def test_finalizers_simple(self):
        t = self.Task(job=sum,
                      args=((1, 2, 3), 0))

        self.assertEqual(t.finalizers, [])

        t.finalizers = [
            Task(job=sum,
                 args=((4, 5, 6, i), 0))
            for i in range(10)
        ]

        self.assertEqual(len(t.finalizers), 10)
        self.assertEqual(t.run(), 6)
        self.assertEqual(len(t.finalizers), 10)
        finals = t.finalize()
        self.assertEqual(len(finals), 10)
        self.assertEqual(len(t.finalizers), 10)
        self.assertEqual(finals,
                         [15, 16, 17, 18, 19, 20, 21, 22, 23, 24])

    def test_finalizers_nested(self):
        t = self.Task(job=sum,
                      args=((1, 2, 3), 0))

        self.assertEqual(t.finalizers, [])

        t.finalizers = [
            Task(job=sum,
                 args=((4, 5, 6, i), 0))
            for i in range(10)
        ]
        self.assertEqual(len(t.finalizers), 10)

        for task in t.finalizers:
            task.finalizers.extend([
                Task(job=sum,
                     args=((4, 5, 6, i), 0))
                for i in range(10)
            ])
            self.assertEqual(len(task.finalizers), 10)

        task_result = t.run()
        self.assertEqual(task_result, 6)
        results = [task_result]
        results.extend(t.finalize())

        self.assertEqual(len(results), 111)
        self.assertEqual(results[0], 6)
        self.assertEqual(results[1], 15)

        for result in results:
            self.assertIsInstance(result, numbers.Number)
            self.assertTrue(result >= 6)
            self.assertTrue(result <= 24)


class TestMapTask(BaseTaskSuite, TestCase):
    @classmethod
    def setUp(self):
        self.c = Configuration()
        self.c.runstate = RuntimeStateConfig()
        self.task = MapTask()
        self.task.job = sum
        self.Task = MapTask
        self.task.conf = self.c

    def test_running_map_task(self):
        t = self.Task()
        t.iter = [(i, i + 10) for i in range(10)]
        t.job = sum

        for a, b in zip(t.run(), [10, 12, 14, 16, 18, 20, 22, 24, 26, 28]):
            self.assertEqual(a, b)



# Copyright 2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from giza.content.helper import character_levels
from rstcloth.rstcloth import RstCloth

logger = logging.getLogger('giza.content.steps.views')


def render_steps(steps, conf):
    r = RstCloth()

    header_html = ('<div class="sequence-block">'
                   '<div class="bullet-block">'
                   '<div class="sequence-step">'
                   '{0}'
                   '</div>'
                   '</div>')

    for idx, step in enumerate(steps.ordered_content()):
        step.render()  # run replacements

        if 'number' not in step:
            step.number = idx

        r.directive('only', 'html or dirhtml or singlehtml')
        r.newline()

        r.directive(name='raw',
                    arg='html',
                    content=header_html.format(step.number),
                    indent=3)
        r.newline()

        if 'heading' in step:
            r.heading(text=step.heading,
                      char=character_levels[step.level] * 2,
                      indent=3)
            r.newline()

        render_step_content(step, 3, r)

        r.directive(name='raw',
                    arg='html',
                    content="</div>",
                    indent=3)
        r.newline()

        r.directive('only', 'not(html or dirhtml or singlehtml)')
        r.newline()

        if 'heading' in step:
            r.heading(text="Step {0}: {1}".format(step.number, step.heading),
                      char=character_levels[step.level] * 2,
                      indent=3)
            r.newline()
        else:
            r.heading(text="Step {0}".format(step.number),
                      char=character_levels[step.level] * 2,
                      indent=3)
            r.newline()

        render_step_content(step, 3, r)

    return r


def render_step_content(step, indent, r):
    if 'pre' in step:
        r.content(content=step.pre,
                  indent=indent,
                  wrap=False)
        r.newline()

    if 'action' in step:
        for action in step.action:
            action.replacement = step.replacement
            action.render()
            render_action(action, indent, step.level + 1, r)

    if 'content' in step:
        r.content(content=step.content,
                  indent=indent,
                  wrap=False)
        r.newline()

    if 'post' in step:
        r.content(content=step.post,
                  indent=indent,
                  wrap=False)
        r.newline()


def render_action(action, indent, level, r):
    if 'heading' in action:
        if level in ('title', 0, 1):
            r.titleg(text=action.heading,
                     char=character_levels[level],
                     indent=indent)
        else:
            r.heading(text=action.heading,
                      char=character_levels[level],
                      indent=indent)
        r.newline()

    if 'pre' in action:
        r.content(content=action.pre,
                  indent=indent,
                  wrap=False)
        r.newline()

    if 'code' in action:
        options = {}
        if action.copyable == False:
            options['fields'] = [('copyable', 'false')]

        r.directive(name='code-block',
                    arg=action.language,
                    indent=indent,
                    content=action.code,
                    wrap=False,
                    **options)
        r.newline()

    if 'content' in action:
        r.content(content=action.content,
                  indent=indent,
                  wrap=False)
        r.newline()

    if 'post' in action:
        r.content(content=action.post,
                  indent=indent,
                  wrap=False)
        r.newline()



# Copyright 2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import shutil

from giza.libgiza.task import Task

from giza.content.steps.inheritance import StepDataCache
from giza.content.steps.views import render_steps
from giza.config.content import new_content_type

logger = logging.getLogger('giza.content.steps.tasks')


def register_steps(conf):
    content_dfn = new_content_type(name='steps',
                                   task_generator=step_tasks,
                                   conf=conf)

    conf.system.content.add(name='steps', definition=content_dfn)


def write_steps(steps, fn, conf):
    content = render_steps(steps, conf)
    content.write(fn)
    logger.debug('wrote steps to: ' + fn)


def step_tasks(conf):
    s = StepDataCache(conf.system.content.steps.sources, conf)
    s.create_output_dir()

    tasks = []
    for fn, stepf in s.file_iter():
        t = Task(job=write_steps,
                 args=(stepf, stepf.target(fn), conf),
                 description='generate a stepfile for ' + fn,
                 target=stepf.target(fn),
                 dependency=fn)
        tasks.append(t)

    logger.debug('added tasks for {0} step generation tasks'.format(len(tasks)))
    return tasks


def step_clean(conf):
    return [Task(job=shutil.rmtree,
                 args=[conf.system.content.steps.output_dir],
                 target=True,
                 dependency=[conf.system.content.steps.output_dir],
                 description='removing {0}'.format(conf.system.content.steps.output_dir))]



# Copyright 2015 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from rstcloth.rstcloth import RstCloth
from rstcloth.table import TableBuilder, RstTable, TableData

logger = logging.getLogger('giza.content.tocs.views')


def render_toctree(toc_items, is_ref=False):
    r = RstCloth()

    r.directive('toctree', fields=[('titlesonly', ''), ('hidden', '')])
    r.newline()

    for entry in toc_items:
        if is_ref is False and 'name' in entry:
            r.content('{0} <{1}>'.format(entry.name, entry.file), indent=3, wrap=False)
        else:
            r.content(entry.file, indent=3, wrap=False)

    return r


def render_dfn_list(toc_items):
    r = RstCloth()

    r.directive('class', 'toc')
    r.newline()

    for entry in toc_items:
        entry.render()
        idnt = 3 * entry.level

        if entry.text_only is True:
            if 'name' in entry:
                r.definition(entry.name, entry.description, indent=idnt)
            else:
                r.content(entry.description, indent=idnt)
            r.newline()
        else:
            if 'name' in entry:
                dfn_heading = r.role('doc', "{0} <{1}>".format(entry.name, entry.file))
            else:
                dfn_heading = r.role('doc', entry.file)

            if 'description' in entry:
                description = entry.description
            else:
                description = ''

            r.definition(dfn_heading, description, indent=idnt)
            r.newline()

    return r


def render_toc_table(toc_items):
    table = TableData()

    table.add_header(['Name', 'Description'])
    for entry in toc_items:
        entry.render()
        if 'name' in entry:
            table.add_row([entry.name, entry.description])
        else:
            table.add_row([RstCloth.role('doc', entry.file), entry.description])

    return TableBuilder(RstTable(table))



import unittest
import tempfile

import giza.libgiza.config


class TestConfigurationObjectPersistance(unittest.TestCase):
    def setUp(self):
        self.conf = giza.libgiza.config.ConfigurationBase

    def test_create_empty_file(self):
        with tempfile.NamedTemporaryFile(suffix=".yaml") as f:
            with self.conf.persisting(f.name) as data:
                data.state["_test_data"] = "42"

            d = self.conf(f.name)

            self.assertEquals(d._test_data, "42")

    def test_operations_with_existing_data(self):
        with tempfile.NamedTemporaryFile(suffix=".yaml") as f:
            with self.conf.persisting(f.name) as data:
                data.state["_test_data"] = 42

            with self.conf.persisting(f.name) as data:
                self.assertEquals(data._test_data, 42)

                data._test_data += 1
                self.assertEquals(data._test_data, 43)

                data.state["_test_data"] = data._test_data

            d = self.conf(f.name)
            self.assertEquals(d._test_data, 43)


class TestConfigurationObjectMembership(unittest.TestCase):
    def setUp(self):
        self.conf = giza.libgiza.config.ConfigurationBase()

    def test_internal_values(self):
        self.assertNotIn("_foo", self.conf)
        self.assertNotIn("_foo", self.conf.state)

        self.conf._foo = 42

        self.assertIn("_foo", self.conf)
        self.assertNotIn("_foo", self.conf.state)



# Copyright 2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
The :mod:`giza.config` stores the models and helpers for all configuration
data. The classes that describe these objects validate input from configuration
files, compute dynamic values, and provide a stable interface for the
configuration data. Important aspects of this code:

:class:`ConfigurationBase` (:mod:`giza.libgiza.config`)
   The base class from which all configuration classes descend. Implements
   :meth:`~ConfigurationBase.ingest()`, which runs after initialization to load
   configuration data from a ``yaml`` file or from a dictionary structure, as
   well as a :meth:`~ConfigurationBase.dict()` method that renders the
   configuration object as a dictionary with some redaction for protected and
   internal fields. Additionally provides customized attribute setting and
   retrieving for these objects.

:class:`RecursiveConfigurationBase` (:mod:`giza.libgiza.base`)
   A base configuration object that contains a reference to the global config
   object in the :attr:`RecursiveConfigurationBase` attribute to facilitate
   resolving values in reference to other values.

:class:`RuntimeStateConfig` (:mod:`giza.config.runtime`)
   By convention, this class defines the ``conf.runstate`` portion of the config
   object, which is populated by the command line argument parsing system. Add
   methods to get and set values, or, add values to the
   :attr:`RuntimeStateConfig._option_registry` to use default setting and
   getting.

:function:`~giza.config.helper.fetch_config()` (:mod:`giza.config.helper`)
   Pass a :class:`RuntimeStateConfig` object to this function, which returns a
   fully constituted configuration object.

:mod:`giza.config.jeerah`
   This module contains all of the configuration classes for the Jira and Github
   (i.e. :mod:`giza.scrumpy` and :mod:`giza.github`) integration.
"""



# Copyright 2015 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import sys

from giza.inheritance import InheritableContentBase

logger = logging.getLogger('giza.content.glossary.models')

if sys.version_info >= (3, 0):
    basestring = str


class GlossaryData(InheritableContentBase):
    @property
    def term(self):
        return self.state['term']

    @term.setter
    def term(self, value):
        if isinstance(value, basestring):
            self.state['term'] = value
        else:
            raise TypeError

    @property
    def definition(self):
        return self.state['definition']

    @definition.setter
    def definition(self, value):
        if isinstance(value, basestring):
            self.state['definition'] = value
        else:
            raise TypeError



# Copyright 2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
:mod:`app` holds the :class:`~giza.app.BuildApp()` class that provides an
organizing framework for running larger sequences of operations.
"""

import collections
import contextlib
import logging
import random
import numbers

import giza.libgiza.pool

from giza.libgiza.task import Task, MapTask
from giza.libgiza.config import ConfigurationBase

logger = logging.getLogger('giza.libgiza.app')


class BuildApp(object):
    """
    A unit of work. :class:`~giza.app.BuildApp()` instances possess queues of
    :class:`~giza.task.Task()` objects and sub-\ :class:`~giza.app.BuildApp()`
    objects that describe a build process. Groups of :class:`~giza.task.Task()`
    objects may execute in parallel, while any :class:`~giza.app.BuildApp()`
    objects in the queue execute in isolation after proceeding group of
    :class:`~giza.task.Task()` operations complete. Mix
    :class:`~giza.app.BuildApp()` and :class:`~giza.task.Task()` operations to
    control task ordering.

    The results of all operations are accessible in the
    :attr:`~giza.app.BuildApp.results`, which largely preserves the ordering of
    the insertion of operations into the queue. Unlike the queue,
    :attr:`~giza.app.BuildApp.results` contains the result of each operation in
    an embedded :class:`~giza.app.BuildApp()` in the order that each task was
    added to the embedded :class:`~giza.app.BuildApp()` instance.

    :class:`~giza.app.BuildApp()` are reusable: after running all operations in
    the queue, the queue resets. However, results do not reset.
    """

    def __init__(self, conf=None, force=False):
        """
        :param ConfigurationBase conf: A top level
           ``Configuration`` object.
        """

        self._conf = conf
        self._force = force
        self._default_pool = 'lazy'
        self._pool_size = None

        self.queue = []
        self.results = []
        self.worker_pool = None
        self._randomize = False

        self.pool_mapping = {
            'thread': giza.libgiza.pool.ThreadPool,
            'process': giza.libgiza.pool.ProcessPool,
            'event': giza.libgiza.pool.EventPool,
            'serial': giza.libgiza.pool.SerialPool
        }

        self.pool_types = tuple(self.pool_mapping.values())

        # so that apps are compatible with tasks in pools
        self.needs_rebuild = True
        self.finalizers = []

        self.root_app = True

        self.target = None
        self.dependency = None

    @classmethod
    def new(cls, pool_type='process', pool_size=None, force=False):
        app = cls()
        app.force = force
        app.default_pool = pool_type
        app.pool_size = pool_size

        return app

    @property
    def dependency(self):
        return self._dependency

    @dependency.setter
    def dependency(self, value):
        self._dependency = value

    @property
    def target(self):
        return self._target

    @target.setter
    def target(self, value):
        self._target = value

    @property
    def description(self):
        jobs = str([j.job if isinstance(j, Task) else type(j)
                    for j in self.queue])

        if self.root_app is True:
            return "a root level BuildApp object: " + jobs
        else:
            return "a BuildApp member BuildApp object: " + jobs

    @property
    def force(self):
        if self._force is None:
            if self.conf is None:
                self._force = False
            else:
                logger.warning('deprecated use of conf object in app setup for force value')
                self._force = self.conf.runstate.force

        return self._force

    @force.setter
    def force(self, value):
        self._force = bool(value)

    @property
    def pool_size(self):
        if self._pool_size is None:
            if self.conf is None:
                # the pool objects themselves know what to do, and so we don't
                # need to over-define defaults here.
                return None
            else:
                logger.warning('deprecated use of conf object for setting pool size')
                self.pool_size = self.conf.runstate.pool_size

        return self._pool_size

    @pool_size.setter
    def pool_size(self, value):
        if isinstance(value, numbers.Number):
            self._pool_size = value
        else:
            logger.warning('{0} is an invalid pool size'.format(str(value)))

    @property
    def conf(self):
        return self._conf

    @conf.setter
    def conf(self, value):
        if isinstance(value, ConfigurationBase):
            self._conf = value

    @property
    def randomize(self):
        return self._randomize

    @randomize.setter
    def randomize(self, value):
        if isinstance(value, bool):
            self._randomize = value
        else:
            self._randomize = bool(value)

    @property
    def default_pool(self):
        if self._default_pool is None:
            if self.conf is None:
                logger.warning('pool type not specified, choosing at random')
                self.default_pool = 'random'
            else:
                logger.warning('deprecated use of conf object in app setup for pool type')
                self._default_pool = self.conf.runstate.runner

        if self.root_app is True and self._default_pool in (None, 'lazy'):
            self.default_pool = 'random'

        return self._default_pool

    @default_pool.setter
    def default_pool(self, value):
        if value == 'lazy':
            pass
        elif value == 'random':
            self._default_pool = random.choice(['process', 'thread', 'serial'])
        elif value in self.pool_mapping:
            self._default_pool = value
        else:
            logger.error('{0} is not a valid pool type'.format(value))

    def define_dependency_node(self, target, dependency):
        self.target = target
        self.dependency = dependency

    def reset(self):
        self.randomize = False
        self.queue = []
        self.results = []

    @property
    def pool(self):
        return self.worker_pool

    @pool.setter
    def pool(self, value=None):
        if value == self.worker_pool:
            pass
        elif isinstance(value, self.pool_types):
            self.worker_pool = value
        elif value in self.pool_mapping:
            self.default_pool = value
        elif value in self.pool_types:
            self.worker_pool = value(self.pool_size)

    def create_pool(self, pool=None):
        if isinstance(pool, self.pool_types):
            self.pool = pool
            return
        elif self.has_active_pool():
            logger.debug('pool exists, not creating a new pool')
            return
        elif pool in self.pool_mapping:
            self.default_pool = pool
        elif self.default_pool == 'lazy' or pool == 'lazy':
            logger.debug('avoiding creating a lazy pool')
            return
        elif pool is None:
            pool = self.default_pool
        elif pool not in self.pool_mapping:
            m = '{0} is not a valid pool type, using the default: {1}'
            logger.error(m.format(pool, self.default_pool))
            pool = self.default_pool

        if self.root_app is False:
            logger.warning('creating a worker pool on a sub_app is probably an error.')

        self.pool = self.pool_mapping[pool](self.pool_size)

    def has_active_pool(self):
        if isinstance(self.worker_pool, self.pool_types):
            return True
        else:
            return False

    @property
    def queue_has_apps(self):
        for task in self.queue:
            if isinstance(task, BuildApp):
                return True

        return False

    def clean_queue(self):
        old_queue_len = len(self.queue)

        new_queue = []
        for task in self.queue:
            if isinstance(task, BuildApp):
                if len(task.queue) == 0:
                    logger.warning('dropping an empty app from task queue')
                    continue
                else:
                    if task.pool is None:
                        task.pool = self.pool
                    new_queue.append(task)
            elif isinstance(task, Task):
                new_queue.append(task)

        self.queue = new_queue

        if len(self.queue) < old_queue_len:
            logger.warning('cleansed queue of empty apps')

    def close_pool(self):
        if self.has_active_pool():
            self.worker_pool.close()
            self.worker_pool = None

    def sub_app(self):
        app = BuildApp()
        app.force = self.force
        app.root_app = False
        app.default_pool = self.default_pool
        app.pool = self.pool

        if self.conf is not None:
            app.conf = self.conf

        return app

    def extend_queue(self, tasks):
        if isinstance(tasks, (Task, BuildApp)):
            self.add(tasks)
        elif tasks is None or len(tasks) == 0:
            return
        else:
            for task in tasks:
                if isinstance(task, collections.Iterable):
                    if len(task) == 0:
                        continue
                    else:
                        app = self.sub_app()
                        app.extend_queue(task)
                        self.add(app)
                else:
                    self.add(task)

    def add(self, task=None, conf=None):
        """
        Adds a new :class:`~giza.app.BuildApp()` or :class:`~giza.task.Task()`
        to the :class:`~giza.app.BuildApp()` object.

        :param string,Task,BuildApp task: Optional. If not specified,
           :meth:`~giza.app.BuildApp.add()` creates and returns a new
           :class:`~giza.task.Task()` object. You can pass the string ``task``
           or the class :class:`~giza.task.Task` to explicitly create a new
           Task, or pass an existing :class:`~giza.task.Task()` instance to add
           that task to the :class:`~giza.app.BuildApp()` instance. You can
           also pass the string ``app`` or the :class:`~giza.app.BuildApp`
           class, to create and add new :class:`~giza.app.BuildApp()`: pass an
           existing :class:`~giza.app.BuildApp()` instance to add that that
           operation grouping to the queue.

        :returns: A reference to a :class:`~giza.app.BuildApp()` or
           :class:`~giza.task.Task()` object in the :class:`~giza.app.BuildApp()`

        :raises: :exc:`TypeError` if the ``task`` argument is invalid.

        """
        if conf is not None:
            self.conf = conf

        if task is None or task in (Task, 'task'):
            t = Task()
            t.conf = self.conf
            t.force = self.force
            self.queue.append(t)
            return t
        elif task in (MapTask, 'map'):
            t = MapTask()
            t.conf = self.conf
            t.force = self.force
            self.queue.append(t)
            return t
        elif task in (BuildApp, 'app'):
            t = self.sub_app()
            self.queue.append(t)
            return t
        else:
            if isinstance(task, Task):
                task.force = self.force
                if task.conf is None:
                    task.conf = self.conf

                self.queue.append(task)
                return task
            elif isinstance(task, BuildApp):
                task.root_app = False
                task.defualt_pool = self.default_pool
                task.force = self.force
                task.pool = self.pool
                self.queue.append(task)
                return task
            else:
                raise TypeError('invalid task type')

    def _run_mixed_queue(self):
        group = []

        for task in self.queue:
            if isinstance(task, Task):
                group.append(task)
            elif isinstance(task, BuildApp):
                if len(group) >= 1:
                    if self.randomize is True:
                        random.shuffle(group)

                    self.results.extend(self.pool.runner(group))
                    group = []

                if task.pool is None:
                    task.pool = self.pool

                self.results.extend(task.run())

        if len(group) != 0:
            self.results.extend(self.pool.runner(group))

    def run(self, randomize=None):
        "Executes all tasks in the :attr:`~giza.app.BuildApp.queue`."

        self.randomize = randomize
        self.create_pool()

        # remove empty apps from queue
        self.clean_queue()

        if len(self.queue) == 0:
            pass  # we could warn here, and while it's not ideal, its mostly harmless.
        elif self.queue_has_apps is True:
            self._run_mixed_queue()
        else:
            if self.randomize is True:
                random.shuffle(self.queue)

            self.results.extend(self.pool.runner(self.queue))

        self.queue = []
        return self.results

    @contextlib.contextmanager
    def context(self, conf=None, randomize=None):
        if len(self.queue) == 0:
            yield self
            self.run(randomize)
        else:
            app = self.add('app')
            yield app
            app.run(randomize)



# Copyright 2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from giza.libgiza.config import ConfigurationBase

logger = logging.getLogger('giza.config.sphinx_local')


class SphinxLocalConfig(ConfigurationBase):
    _option_registry = ['project', 'master_doc', 'logo']

    @property
    def paths(self):
        return self.state['paths']

    @paths.setter
    def paths(self, value):
        if isinstance(value, dict):
            self.state['paths'] = SphinxLocalPaths(value)
        else:
            raise TypeError

    @property
    def theme(self):
        return self.state['theme']

    @theme.setter
    def theme(self, value):
        if isinstance(value, dict):
            self.state['theme'] = SphinxLocalTheme(value)
        else:
            raise TypeError

    @property
    def sidebars(self):
        return self.state['sidebars']

    @sidebars.setter
    def sidebars(self, value):
        if isinstance(value, dict):
            self.state['sidebars'] = value
        else:
            raise TypeError


class SphinxLocalPaths(ConfigurationBase):
    _option_registry = ['static', 'locale']


class SphinxLocalTheme(ConfigurationBase):
    _option_registry = ['name', 'project', 'google_analytics', 'book_path_base',
                        'repo', 'jira', 'sitename']

    @property
    def nav_excluded(self):
        return self.state['nav_excluded']

    @nav_excluded.setter
    def nav_excluded(self, value):
        if isinstance(value, list):
            for path in value:
                if not path.startswith('/'):
                    raise TypeError

            self.state['nav_excluded'] = value
        else:
            raise TypeError



# Copyright 2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from rstcloth.rstcloth import RstCloth

logger = logging.getLogger('giza.content.options.views')


def render_options(option, conf):
    r = RstCloth()

    if 'program' not in option.replacement:
        option.replacement['program'] = RstCloth.role('program', option.program)

    if option.has_field('command'):
        option.replacement['command'] = RstCloth.role('toolcommand', option.command)

    if option.directive in ['option']:
        if len(option.name) > 1 and option.name[0] in ('<', '-'):
            prefix = ''
        else:
            prefix = '--'

        directive_str = '{prefix}{name}'

        if option.has_field('args'):
            directive_str += ' {args}'

        if option.has_field('aliases'):
            directive_str += ', {0}'.format(option.aliases)

            if option.has_field('args'):
                directive_str += ' {args}'

        if option.has_field('args'):
            directive_str = directive_str.format(prefix=prefix, name=option.name, args=option.args)
        else:
            directive_str = directive_str.format(prefix=prefix, name=option.name)
    else:
        prefix = ''
        directive_str = option.name

    if 'role' not in option.replacement:
        option.replacement['role'] = ':{0}:`{1}{2}`'.format(option.directive, prefix, option.name)

    option.render()  # jinja template render
    r.directive(option.directive, directive_str)
    r.newline()

    indent = 3
    if option.has_field('type'):
        r.content('*Type*: {0}'.format(option.type), indent=indent)
        r.newline()

    if option.has_field('default'):
        r.content('*Default*: {0}'.format(option.default), indent=indent)
        r.newline()

    for field in ('pre', 'description', 'content', 'post'):
        if option.has_field(field) is False:
            continue
        else:
            r.content(getattr(option, field).split('\n'), indent=indent, wrap=False)
            r.newline()

    return r



"""
A collection of tests of the TypedDict specialized dictionary type which allows
us to specify types for of the the keys and values of a dictionary along with
validation methods to validate input on creation.

These tests ensure that TypedDicts behave as expected, which is to say, like
"vanilla" dicts, but that they also enforce their validation requirements (when
implemented.) and type checking (always.)
"""
import datetime
import sys
import unittest

import giza.libgiza.typed_dict
import giza.libgiza.error

if sys.version_info >= (3, 0):
    basestring = str


class TestTypedDictionaryObjectCreation(unittest.TestCase):
    def test_object_creation_returns_correctly_typed_values(self):
        d = giza.libgiza.typed_dict.TypedDict(key_type=basestring,
                                              value_type=bool)

        self.assertIsInstance(d, giza.libgiza.typed_dict.TypedDict)

        d = giza.libgiza.typed_dict.TypedDict(key_type=basestring,
                                              value_type=bool)

        self.assertIsInstance(d, giza.libgiza.typed_dict.TypedDict)

    def test_object_initialization_with_invalid_inputs(self):
        with self.assertRaises(TypeError):
            giza.libgiza.typed_dict.TypedDict()

        with self.assertRaises(TypeError):
            giza.libgiza.typed_dict.TypedDict("1", 2)

        with self.assertRaises(TypeError):
            giza.libgiza.typed_dict.TypedDict(bool, 2)

        with self.assertRaises(TypeError):
            giza.libgiza.typed_dict.TypedDict(2, bool)

        with self.assertRaises(TypeError):
            giza.libgiza.typed_dict.TypedDict({})

        with self.assertRaises(TypeError):
            giza.libgiza.typed_dict.TypedDict("1", 2, {})

        with self.assertRaises(TypeError):
            giza.libgiza.typed_dict.TypedDict(bool, 2, {})

        with self.assertRaises(TypeError):
            giza.libgiza.typed_dict.TypedDict(2, bool, {})

    def test_check_persistence_of_values_set_after_creation(self):
        d = giza.libgiza.typed_dict.TypedDict(basestring, bool)
        d["foo"] = True
        self.assertEquals(True, d["foo"])

    def test_abc_implementations_return_correctly_typed_values(self):
        d = giza.libgiza.typed_dict.TypedDict(basestring, bool)

        self.assertIsInstance(d.check_key("foo"), giza.libgiza.error.ErrorCollector)
        self.assertIsInstance(d.check_value(True), giza.libgiza.error.ErrorCollector)
        self.assertIsInstance(d.check_pair("foo", True), giza.libgiza.error.ErrorCollector)

    def test_creation_of_objects_works_correctly_with_input_base_object(self):
        for base in [{"foo": True, "bar": False}, [("foo", True), ("bar", False)],
                     {"baz": False}, [("baz", False)], {}, [], tuple()]:
            d = giza.libgiza.typed_dict.TypedDict(basestring, bool)
            d.ingest(base)
            self.assertTrue(len(d) == len(base))


class Fake(object):
    def __init__(self, left, right):
        self.value = (left, right)
        self.validate_results = giza.libgiza.error.ErrorCollector()

    def validate(self):
        return self.validate_results


class FakeTypedDict(giza.libgiza.typed_dict.TypedDict):
    def __init__(self, *args):
        super(FakeTypedDict, self).__init__(key_type=Fake,
                                            value_type=Fake)
        self.ingest(args)
        self.pair_results = giza.libgiza.error.ErrorCollector()

    def check_key(self, key):
        collector = key.validate()
        if collector.has_errors() and collector.fatal:
            raise ValueError("key {0} has an error".format(key))

        return collector

    def check_value(self, value):
        collector = value.validate()
        if collector.has_errors() and collector.fatal:
            raise ValueError
        else:
            return collector

    def check_pair(self, key, value):
        if self.pair_results.has_errors() and self.pair_results.fatal:
            raise ValueError
        else:
            return self.pair_results


class TestTypedDictionaryOperations(unittest.TestCase):
    def setUp(self):
        self.d = FakeTypedDict()
        self.key = Fake(1, 2)
        self.value = Fake(2, 3)

    def test_fake_object_exists_with_correct_types(self):
        self.assertIsInstance(self.d, FakeTypedDict)
        self.assertIsInstance(self.d, giza.libgiza.typed_dict.TypedDict)
        self.assertIsInstance(self.d, dict)

    def test_setting_object_invalid_types_raises_type_errors(self):
        self.assertTrue(len(self.d) == 0)

        pairs = [
            (1, 1),
            (True, False),
            (1, True),
            (False, 0),
            (None, True),
            (False, None),
            ("string value", 3),
            (4, "string value"),
            (Fake(1, 2), 3),
            (Fake(2, 3), "3"),
            (Fake(4, 5), True),
            (3, Fake(5, 6)),
            ("4", Fake(6, 7)),
            (None, Fake(7, 8))
        ]

        for k, v in pairs:
            with self.assertRaises(TypeError):
                self.d[k] = v

        self.assertTrue(len(self.d) == 0)

    def test_setting_object_valid_data_allows_recall(self):
        self.d[self.key] = self.value

        self.assertIs(self.value, self.d[self.key])
        self.assertTrue(len(self.d) == 1)

    def test_setting_object_with_invalid_key_raises_value_error(self):
        self.key.validate_results.add(giza.libgiza.error.Error(message="key has an error"))

        with self.assertRaises(ValueError):
            self.d[self.key] = self.value

    def test_setting_object_with_invalid_value_raises_value_error(self):
        self.key.validate_results.add(giza.libgiza.error.Error(message="value has an error"))

        with self.assertRaises(ValueError):
            self.d[self.key] = self.value

    def test_setting_object_with_invalid_value_and_key_raises_value_error(self):
        self.key.validate_results.add(giza.libgiza.error.Error(message="key has an error"))
        self.key.validate_results.add(giza.libgiza.error.Error(message="value has an error"))

        with self.assertRaises(ValueError):
            self.d[self.key] = self.value

    def test_setting_with_invalid_pair_raises_value_error(self):
        self.key.validate_results.add(giza.libgiza.error.Error(message="an object has errors"))

        with self.assertRaises(ValueError):
            self.d[self.key] = self.value

    def test_abc_implementations_of_checks_return_correctly_typed_values(self):
        self.assertIsInstance(self.d.check_key(self.key), giza.libgiza.error.ErrorCollector)
        self.assertIsInstance(self.d.check_value(self.value), giza.libgiza.error.ErrorCollector)
        self.assertIsInstance(self.d.check_pair(self.key, self.value),
                              giza.libgiza.error.ErrorCollector)

    def test_check_functions_raise_exception(self):
        def bad_pair_validator(self, key, value):
            raise AttributeError("error")

        self.d.check_pair = bad_pair_validator

        with self.assertRaises(ValueError):
            self.d[self.key] = self.value


class LatestReleaseDownloads(giza.libgiza.typed_dict.TypedDict):
    def __init__(self, *args):
        super(LatestReleaseDownloads, self).__init__(key_type=basestring,
                                                     value_type=basestring)
        self.ingest(args)

    def check_key(self, key):
        return giza.libgiza.error.ErrorCollector()

    def check_value(self, value):
        errors = giza.libgiza.error.ErrorCollector()

        if "win" in value:
            if not value.endswith(".zip"):
                errors.add(giza.libgiza.error.Error(
                    message="windows binaries must end with .zip, not: " + value))
        else:
            if not value.endswith(".tgz"):
                errors.add(giza.libgiza.error.Error(
                    message="unix-like packages should end with .tgz: " + value))

        return errors

    def check_pair(self, key, value):
        errors = giza.libgiza.error.ErrorCollector()

        if key.replace("_", "-", 1) not in value:
            errors.add(giza.libgiza.error.ErrorCollector(
                message="key '{0}' is not in value '{1}'".format(key, value)))

        return errors


class LatestReleaseDict(giza.libgiza.typed_dict.TypedDict):
    def __init__(self, *args):
        super(LatestReleaseDict, self).__init__(key_type=basestring,
                                                value_type=LatestReleaseDocument)
        self.ingest(args)

    def check_key(self, key):
        return None

    def check_value(self, value):
        return value.validate()

    def check_pair(self, key, value):
        return None


class LatestReleaseDocument(giza.libgiza.config.ConfigurationBase):
    _option_registry = ["major", "minor", "maintenance"]

    @property
    def version(self):
        return self.state["version"]

    @version.setter
    def version(self, value):
        self.state["version"] = value

    @property
    def date(self):
        return self.state["date"]

    @date.setter
    def date(self, value):
        date_format = "%m/%d/%Y"
        if isinstance(value, datetime.datetime):
            self.state["date"] = value.strftime(date_format)
        elif isinstance(value, basestring):
            # pass through a datetime time object to make sure its valid and
            # well formed.
            self.state["date"] = datetime.datetime.strptime(value,
                                                            date_format).strftime(date_format)
        else:
            raise TypeError("invalid date value: {0}".format(value))

    @property
    def rc(self):
        if "rc" not in self.state:
            return False
        else:
            return self.state["rc"]

    @rc.setter
    def rc(self, value):
        if isinstance(value, bool):
            self.state["rc"] = value
        else:
            raise TypeError("{0} is not a valid 'rc' value".format(value))

    @property
    def downloads(self):
        if "downloads" not in self.state:
            self.state["downloads"] = LatestReleaseDownloads()

        return self.state["downloads"]

    @downloads.setter
    def downloads(self, value):
        self.state["downloads"] = LatestReleaseDownloads(value)

    def validate(self):
        errors = giza.libgiza.error.ErrorCollector()

        for key in ("rc", "version", "major", "minor", "maintenance", "date", "downloads"):
            if key not in self.state:
                m = ("missing {0} in latest release for series {1} "
                     "object".format(key, self.version))
                errors.add(giza.libgiza.error.Error(message=m, include_trace=False))

        for platform, url in self.downloads.items():
            errors.add(self.downloads.check_pair(platform, url))
            errors.add(self.downloads.check_key(platform))
            errors.add(self.downloads.check_value(url))

            if self.version not in url:
                m = ("incorrect version ({0}) for url "
                     "{1}".format(self.version.string, url))
                errors.add(giza.libgiza.error.Error(message=m, include_trace=False))

        return errors


class TestLatestReleaseTypedDict(unittest.TestCase):
    def setUp(self):
        self.d = LatestReleaseDict()
        self.value = LatestReleaseDocument()
        self.value.version = "2.2.2"
        self.value.major = 2
        self.value.minor = 2
        self.value.maintenance = 2
        self.value.rc = False
        self.value.date = "1/16/2014"
        self.value.downloads = {"foo": "https://fastdl.mongodb.org/foo/mongodb-2.2.2-foo.tgz"}

    def test_adding_version(self):
        self.d["2.2"] = self.value
        self.assertIs(self.value, self.d["2.2"])

    def test_creation_from_existing_object(self):
        self.d["2.2"] = self.value

        o = LatestReleaseDict(self.d)
        self.assertEquals(o["2.2"], self.d["2.2"])

    def test_adding_version_with_missing_value(self):
        del self.value.state["date"]

        with self.assertRaises(ValueError):
            self.d["2.2"] = self.value



# Copyright 2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os.path

import yaml

from giza.libgiza.config import RecursiveConfigurationBase, ConfigurationBase
from giza.config.sphinx_local import SphinxLocalConfig
from giza.config.manpage import ManpageConfig
from giza.config.pdfs import PdfConfig
from giza.config.intersphinx import IntersphinxConfig
from giza.config.redirects import HtaccessData
from giza.config.content import ContentRegistry
from giza.config.replacements import ReplacementData
from giza.config.migrations import MigrationData
from giza.config.images import ImageData
from giza.config.jeerah import JeerahConfig

logger = logging.getLogger('giza.config.system')


class SystemConfig(RecursiveConfigurationBase):

    @property
    def make(self):
        return self.state['make']

    @make.setter
    def make(self, value):
        self.state['make'] = SystemMakeConfig(value)

    @property
    def tools(self):
        return self.state['tools']

    @tools.setter
    def tools(self, value):
        self.state['tools'] = SystemToolsConfig(value)

    @property
    def files(self):
        if 'files' not in self.state:
            self.files = []

        return self.state['files']

    @files.setter
    def files(self, value):
        if 'files' not in self.state:
            self.state['files'] = SystemConfigFiles(self.conf)
            self.files.paths = value

    @property
    def conf_file(self):
        if 'conf_file' not in self.state:
            self.conf_file = None

        return self.state['conf_file']

    @conf_file.setter
    def conf_file(self, value):
        pass

    @property
    def branched(self):
        return self.conf.project.branched

    @branched.setter
    def branched(self, value):
        logger.warning('branched state is stored in system not project.')

    @property
    def dependency_cache(self):
        if 'dependency_cache' not in self.state:
            self.dependency_cache = None

        return self.state['dependency_cache']

    @dependency_cache.setter
    def dependency_cache(self, value):
        if value is not None:
            self.state['dependency_cache'] = value
        else:
            self.state['dependency_cache'] = os.path.join(self.conf.paths.projectroot,
                                                          self.dependency_cache_fn)

    @property
    def dependency_cache_fn(self):
        if 'dependency_cache_fn' not in self.state:
            self.dependency_cache_fn = None

        return self.state['dependency_cache_fn']

    @dependency_cache_fn.setter
    def dependency_cache_fn(self, value):
        if value is None:
            p = [self.conf.paths.branch_output]
            if self.conf.project.edition is None:
                p.append('dependencies.json')
            else:
                p.append('dependencies-' + self.conf.project.edition + '.json')

            self.state['dependency_cache_fn'] = os.path.sep.join(p)
        else:
            self.state['dependency_cache_fn'] = value

    @property
    def runstate(self):
        return self.conf.runstate

    @runstate.setter
    def runstate(self, value):
        self.conf.runstate = value

    @property
    def content(self):
        if 'content' not in self.state:
            self.state['content'] = ContentRegistry()

        return self.state['content']


class SystemToolsConfig(ConfigurationBase):

    @property
    def pinned(self):
        return self.state['pinned']

    @pinned.setter
    def pinned(self, value):
        if isinstance(value, bool):
            self.state['pinned'] = value
        else:
            raise TypeError

    @property
    def ref(self):
        return self.state['ref']

    @ref.setter
    def ref(self, value):
        if value in ('HEAD', 'master') or len(value) == 40:
            self.state['ref'] = value
        else:
            raise TypeError


class SystemMakeConfig(ConfigurationBase):

    @property
    def generated(self):
        return self.state['generated']

    @generated.setter
    def generated(self, value):
        if isinstance(value, list):
            self.state['generated'] = value
        else:
            raise TypeError

    @property
    def static(self):
        return self.state['static']

    @static.setter
    def static(self, value):
        if isinstance(value, list):
            self.state['static'] = value
        else:
            raise TypeError


class SystemConfigFiles(RecursiveConfigurationBase):

    def __init__(self, conf):
        super(SystemConfigFiles, self).__init__(None, conf)

    @property
    def paths(self):
        return self.state['paths']

    @paths.setter
    def paths(self, value):
        if isinstance(value, list):
            self.state['paths'] = value
        else:
            raise TypeError

    @property
    def data(self):
        if 'data' not in self.state:
            self.data = self.conf.system.files.paths

        return self.state['data']

    @data.setter
    def data(self, value):
        if 'data' not in self.state:
            self.state['data'] = SystemConfigData(value, self.conf)

    def get_configs(self, key):
        fns = []
        for value in self.conf.system.files.paths:
            if isinstance(value, dict):
                if 'migration' in value:
                    if isinstance(value[key], list):
                        fns.extend(value[key])
                    else:
                        fns.append(value[key])
            elif value.startswith(key):
                fns.append(value)

        results = []

        for fn in fns:
            if fn.startswith(os.path.sep):
                fn = fn[len(os.path.sep):]

            for new_file in [os.path.join(self.conf.paths.projectroot, fn),
                             os.path.join(self.conf.paths.projectroot,
                                          self.conf.paths.source, fn),
                             os.path.join(self.conf.paths.projectroot,
                                          self.conf.paths.builddata, fn)]:
                if os.path.isfile(new_file):
                    results.append(new_file)

        return results


class SystemConfigData(RecursiveConfigurationBase):
    # There shouldn't be any setters in this class. All items in this class
    # must exist in SystemConfigPaths() objects.

    _always_list_configs = ['errors', 'images', 'intersphinx', 'manpages',
                            'pdfs', 'push', 'robots']
    _single_document_configs = ['errors']

    def __init__(self, obj, conf):
        super(SystemConfigData, self).__init__(None, conf)
        for fn in self.conf.system.files.paths:
            if isinstance(fn, dict):
                attr_name = next(iter(fn))
            else:
                attr_name = os.path.splitext(fn)[0]

            self._option_registry.append(attr_name)
            setattr(self, attr_name, fn)

    def __getattr__(self, key):
        try:
            return object.__getattribute__(self, key)
        except AttributeError:
            pass

        if key not in self._option_registry:
            m = 'key "{0}" in system.data object does not exist'.format(key)
            if not key.startswith('__'):
                logger.debug(m)
            raise AttributeError(m)

        if isinstance(key, dict):
            basename, fn = next(iter(key.items()))
        else:
            fn = key
            basename = os.path.splitext(fn)[0]

        if isinstance(self.state[key], list):
            pass
        elif isinstance(self.state[key], dict):
            if len(self.state[key]) == 1:
                basename, fns = next(iter(self.state[key].items()))
                if isinstance(fns, list):
                    for fn in fns:
                        self._load_file(basename, fn)
                else:
                    self._load_file(basename, fn)
        else:
            self._load_file(basename, self.state[key])

        if len(self.state[key]) == 1 and key not in self._always_list_configs:
            return self.state[key][0]
        else:
            return self.state[key]

    def __contains__(self, value):
        return value in self._option_registry

    def ingest(self, file_list=None):
        # this needs to be a noop, because some of the base constructors call
        # it, but we want this class to load data lazily.
        pass

    def _load_file(self, basename, fn):
        full_path = self._resolve_config_path(fn)

        if basename not in self.state:
            if os.path.isfile(full_path):
                # TODO we should make this process lazy with a more custom getter/setter
                self.state[basename] = self._resolve_config_data(full_path, basename)
                logger.debug('set sub-config %s with data from %s', basename, full_path)
            else:
                self.state[basename] = []
                logger.warning('{0} does not exist. continuing.'.format(full_path))
        else:
            if os.path.isfile(full_path):
                d = self._resolve_config_data(full_path, basename)
            else:  # if os.path.isfile(self.state[basename]):
                full_path = self.state[basename]  # .items()[0][1]
                d = self._resolve_config_data(self._resolve_config_path(full_path), basename)
            self._set_config_data(basename, fn, d)

    def keys(self):
        return self._option_registry

    def _set_config_data(self, basename, fn, d):
        if isinstance(self.state[basename], list):
            if isinstance(d, list):
                self.state[basename].extend(d)
            else:
                self.state[basename].append(d)
        else:
            if (isinstance(self.state[basename], (ConfigurationBase, dict)) and
                    len(self.state[basename]) == 1):

                self.state[basename] = []
            elif self.state[basename] != fn:
                self.state[basename] = [self.state[basename]]
            else:
                self.state[basename] = []

            if isinstance(d, list):
                self.state[basename].extend(d)
            else:
                self.state[basename].append(d)

    def _resolve_config_path(self, fn):
        if isinstance(fn, dict):
            if len(fn) == 1:
                fn = next(iter(fn.values()))
            else:
                logger.error("unsupported config file specification: " + str(fn))

        if os.path.exists(fn):
            full_path = fn
        elif os.path.exists(os.path.join(os.getcwd(), fn)):
            full_path = os.path.join(os.getcwd(), fn)
        elif fn.startswith('/'):
            full_path = os.path.join(self.conf.paths.projectroot, fn[1:])
        else:
            full_path = os.path.join(self.conf.paths.projectroot,
                                     self.conf.paths.builddata, fn)

        return full_path

    def _resolve_config_data(self, fn, basename):
        logger.debug('resolving config data from file ' + fn)
        if fn is None:
            return []
        else:
            mapping = {
                'sphinx_local': SphinxLocalConfig,
                'sphinx-local': SphinxLocalConfig,
                'manpages': ManpageConfig,
                'pdfs': PdfConfig,
                'intersphinx': IntersphinxConfig,
            }
            # recur_mapping for config objects that subclass RecursiveConfigurationBase
            recur_mapping = {
                'jira': JeerahConfig
            }
            special_lists = {
                'htaccess': HtaccessData,
                'migrations': MigrationData,
                'images': ImageData,
            }
            self._always_list_configs.extend(special_lists.keys())

            with open(fn, 'r') as f:
                if basename in self._single_document_configs:
                    data = yaml.safe_load(f)
                else:
                    data = yaml.safe_load_all(f)

                if basename in mapping:
                    data = [mapping[basename](doc) for doc in data]
                elif basename in recur_mapping:
                    data = [recur_mapping[basename](doc, self.conf) for doc in data]
                elif basename in special_lists:
                    l = special_lists[basename]()
                    l.conf = self.conf
                    l.extend([d for d in data])
                    data = l
                elif basename == 'replacement':
                    data = ReplacementData([d for d in data], self.conf)
                    return data

                if not isinstance(data, list):
                    data = [item for item in data]

            if len(data) == 1 and (basename not in self._always_list_configs):
                return data[0]
            else:
                return data



#  2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Contains the output specification for redirects (i.e. ``.htaccess`` files,) as
well as integration into the :class:`giza.libgiza.app.BuildApp()`
infrastructure. All of the data processing and definition happens in
:mod:`giza.config.redirects`.
"""

import os.path
import logging

import giza.libgiza.task

logger = logging.getLogger('giza.content.post.redirects')


def make_redirect(conf):
    o = []

    logger.info('generating {0} redirects'.format(len(conf.system.files.data.htaccess)))
    for redir in conf.system.files.data.htaccess:
        if redir.to.startswith('http'):
            url = redir.to
        else:
            url = conf.project.url + redir.to

        if url.endswith('/'):
            url = url[:-1]

        o.append(' '.join(['Redirect', str(redir.code), redir.from_loc, url, '\n']))

    o.sort()
    o.extend(['\n',
              '<FilesMatch "\.(ttf|otf|eot|woff)$">', '\n',
              '   Header set Access-Control-Allow-Origin "*"', '\n',
              '</FilesMatch>',
              ])

    return o


def write_redirects(fn, conf):
    if not os.path.exists(os.path.dirname(fn)):
        os.makedirs(os.path.dirname(fn))

    with open(fn, 'w') as f:
        f.writelines(make_redirect(conf))
        f.write('\n')

    logger.info('wrote redirects to: ' + fn)


def redirect_tasks(conf):
    tasks = []

    if conf.git.branches.current != 'master':
        return tasks

    if 'htaccess' in conf.system.files.data:
        fn = os.path.join(conf.paths.projectroot, conf.paths.htaccess)

        deps = []
        for configfn in conf.system.files.paths:
            if isinstance(configfn, dict):
                if 'htaccess' in configfn:
                    deps.extend([os.path.join(conf.paths.projectroot, conf.paths.builddata, rfn)
                                 for rfn in configfn['htaccess']])
            elif configfn.startswith('htaccess'):
                deps.append(os.path.join(conf.paths.projectroot, conf.paths.builddata, configfn))

        tasks.append(giza.libgiza.task.Task(job=write_redirects,
                                            args=(fn, conf),
                                            target=fn,
                                            dependency=deps,
                                            description=' '.join(('generate and write redirects into:',
                                                                  conf.paths.htaccess))))

    logger.info("added {0} redirect generation tasks".format(len(tasks)))
    return tasks



"""
Given a a specification of an image, and an ``svg`` in the configured image
directory, build desired image artifacts (using ``inkscape``) as well as an
``rst`` file that includes all of content to include the image in the
output.

With images generated from SVG, we separate the generation of build artifacts
(images) from their source, which means the documentation can use
output-specific images and potentially translate the text in the diagrams. By
generating all of the ``rst`` to inclde the images, its possible display and
maintain the presentation of images systematically and centrally.

The format of the image specification is: ::

  {
    "name": <imgBaseName>,
    "alt": <text>,
    "output": [
      {
        "type": <string>,
        "tag": <string>,
        "dpi": <int>,
        "width": <int>
      },
    ]
  }

The ``tag`` field is optional and appends a string to the generated file name,
to facilitate multiple output targets without having conflicting file names.

The ``<imgBaseName>`` is the name of the ``.svg`` file without the
extension. The main image-generation operation is in
:func:`giza.content.images.image_tasks()`, while definition of the `rst`` content
is in :func:`giza.content.images.generate_image_pages()`.
"""



# Copyright 2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from giza.libgiza.config import RecursiveConfigurationBase

logger = logging.getLogger('giza.config.version')


class VersionConfig(RecursiveConfigurationBase):
    _option_registry = ['release', 'branch']

    def has_version(self, value):
        return ('version' in self.conf.runstate.branch_conf and
                value in self.conf.runstate.branch_conf['version'])

    @property
    def published(self):
        if 'published' not in self.state:
            self.published = None

        return self.state['published']

    @published.setter
    def published(self, value):
        if self.has_version('published'):
            p = self.conf.runstate.branch_conf['version']['published']

            if not isinstance(p, list):
                msg = "published branches must be a list"
                logger.critical(msg)
                raise TypeError(msg)

            self.state['published'] = p
        else:
            self.state['published'] = []

    @property
    def active(self):
        if 'active' not in self.state:
            self.active = None

        return self.state['active']

    @active.setter
    def active(self, value):
        if self.has_version('active'):
            p = self.conf.runstate.branch_conf['version']['active']

            if not isinstance(p, list):
                msg = "active branches must be a list"
                logger.critical(msg)
                raise TypeError(msg)

            self.state['active'] = p
        else:
            self.state['active'] = []

    @property
    def upcoming(self):
        if 'upcoming' not in self.state:
            self.upcoming = None

        return self.state['upcoming']

    @upcoming.setter
    def upcoming(self, value):
        if self.has_version('upcoming'):
            self.state['upcoming'] = self.conf.runstate.branch_conf['version']['upcoming']
        else:
            self.state['upcoming'] = None

    @property
    def stable(self):
        if 'stable' not in self.state:
            self.stable = None

        return self.state['stable']

    @stable.setter
    def stable(self, value):
        if self.has_version('stable'):
            self.state['stable'] = self.conf.runstate.branch_conf['version']['stable']
        else:
            self.state['stable'] = None



# Copyright 2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy


def get_dependency_graph(app):
    graph = {}

    tasks = copy.copy(app.queue)
    for task in tasks:
        if isinstance(task.target, list):
            for target in task.target:
                if target not in graph:
                    graph[target] = []
                if isinstance(task.dependency, list):
                    graph[target].extend(task.dependency)
                else:
                    graph[target].append(task.dependency)
            continue
        elif task.target not in graph:
            graph[task.target] = []

        if task.target is None:
            print(type(task),
                  task.dependency if hasattr(task, 'dependency') else "no dep",
                  task.job if hasattr(task, 'job') else "no job",
                  task.description if hasattr(task, 'description') else "no text")

        if isinstance(task.dependency, list):
            graph[task.target].extend(task.dependency)
        else:
            graph[task.target].append(task.dependency)

    return graph



# Copyright 2015 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy

from rstcloth.rstcloth import RstCloth
from rstcloth.rstcloth import fill
from rstcloth.table import TableData, TableBuilder, ListTable


def setup_replacements(content):
    if content.interface == 'command':
        role_type = 'dbcommand'
    elif content.interface == 'method':
        role_type = 'method'
    elif content.interface == 'phpmethod':
        role_type = 'phpmethod'
    else:
        role_type = 'samp'

    if 'role' not in content.replacement:
        content.replacement['role'] = RstCloth.role(role_type, content.operation)

    if 'type' not in content.replacement:
        if isinstance(content.type, list):
            if len(content.type) == 1:
                content.replacement['type'] = content.type[0]
            elif len(content.type) == 2:
                content.replacement['type'] = 'or'.join(content.type)
            else:
                types = copy.copy(content.type)
                types[-1] = 'and ' + types[-1]
                content.replacement['type'] = ','.join(types)
        else:
            content.replacement['type'] = content.type

    if 'argname' not in content.replacement:
        content.replacement['argname'] = content.name


def render_apiargs(apiargs):
    for content in apiargs.ordered_content():
        setup_replacements(content)
        content.render()  # run_replacements

    r = RstCloth()

    r.directive('only', '(html or singlehtml or dirhtml)')
    r.newline()
    render_apiarg_table(r, apiargs)

    r.newline()

    r.directive('only', 'not(html or singlehtml or dirhtml)')
    r.newline()
    render_apiarg_fields(r, apiargs)

    return r


def render_apiarg_table(r, apiargs):
    table = TableData()

    header = [apiargs.field_type()]

    if apiargs.has_type() is True:
        header.append('Type')

    header.append('Description')

    num_columns = len(header)
    table.add_header(header)

    if num_columns == 2:
        widths = [20, 80]
        for entry in apiargs.ordered_content():
            table.add_row([RstCloth.pre(entry.name),
                           fill(string=entry.description, first=0, hanging=3, wrap=False)])
    elif num_columns == 3:
        widths = [20, 20, 80]
        for entry in apiargs.ordered_content():
            table.add_row([RstCloth.pre(entry.name),
                           entry.type_for_table_output(),
                           fill(string=entry.description, first=0, hanging=3, wrap=False)])

    r.content(TableBuilder(ListTable(table, widths=widths)).output, indent=3)


def render_apiarg_fields(r, apiargs):
    for content in apiargs.ordered_content():

        field_name = [content.arg_name]

        if content.type != '':
            field_name.append(', '.join(content.type))

        field_name.append(content.name)

        field_content = fill(string=content.description,
                             first=0,
                             hanging=6,
                             wrap=False)

        r.field(name=' '.join(field_name),
                value=field_content,
                indent=3,
                wrap=False)
        r.newline()



# Copyright 2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from six.moves.urllib.parse import quote


def equality(user_input):
    return lazy_list(user_input, '=', 'in')


def inequality(user_input):
    return lazy_list(user_input, '!=', 'not in')


def escape_string(string):
    if ' ' in string or '.' in string:
        string = ''.join(['"', string, '"'])
    return string


def lazy_list(user_input, single, multi):
    if not isinstance(user_input, list):
        user_input = [escape_string(a) for a in user_input.split(',')]
    else:
        user_input = [escape_string(s) for s in user_input]

    if len(user_input) <= 1:
        return ' '.join([single, user_input[0]])
    else:
        return ' '.join([multi, '(', ", ".join(user_input), ')'])


def query_link(url, query_string):
    if url.endswith('/'):
        url = url[:-1]
    return '/'.join([url, 'issues', '?jql=' + quote(query_string)])


def strip_name(version):
    if version.startswith('"') and version.endswith('"'):
        version = version[1:-1]

    return version



# Copyright 2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import time

logger = logging.getLogger('giza.tools.timing')


class Timer():

    def __init__(self, name=None):
        if name is None:
            self.name = 'task'
        else:
            self.name = name

    def __enter__(self):
        self.start = time.time()

    def __exit__(self, *args):
        logger.debug('time elapsed for "{0}" was: {1}'.format(
            self.name, str(time.time() - self.start)))



# Copyright 2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os.path
import sys
import logging
import numbers
import collections

import giza.libgiza.config
import giza.content.helper

logger = logging.getLogger('giza.config.images')

if sys.version_info >= (3, 0):
    basestring = str


class ImageData(list):
    def append(self, item):
        self.insert(-1, item)

    def extend(self, items):
        for item in items:
            self.insert(-1, item)

    def insert(self, index, item):
        if item is None:
            return
        elif giza.content.helper.edition_check(item, self.conf) is False:
            return

        super(ImageData, self).insert(index, ImageSpecification(item, self.conf))


class ImageSpecification(giza.libgiza.config.RecursiveConfigurationBase):
    @property
    def name(self):
        return self.state['name']

    @name.setter
    def name(self, value):
        if isinstance(value, basestring):
            self.state['name'] = value
            self.state['source_base'] = os.path.join(self.conf.paths.projectroot,
                                                     self.conf.paths.branch_images, value)
            self.state['source_core'] = os.path.join(self.conf.paths.projectroot,
                                                     self.conf.paths.images, value + '.svg')
        else:
            raise TypeError('{0}: {1}'.format(type(value), value))

    @property
    def alt(self):
        return self.state['alt']

    @alt.setter
    def alt(self, value):
        if isinstance(value, basestring):
            self.state['alt'] = value
        else:
            raise TypeError('{0}: {1}'.format(type(value), value))

    @property
    def outputs(self):
        if 'outputs' in self.state:
            return self.state['outputs']
        else:
            return []

    @outputs.setter
    def outputs(self, value):
        if 'outputs' not in self.state:
            self.state['outputs'] = []

        if isinstance(value, collections.Iterable):
            self.state['outputs'].extend([ImageOutputSpecification(d, self) for d in value])
        else:
            self.state['outputs'].append(ImageOutputSpecification(value, self))

    output = outputs

    @property
    def dir(self):
        return self.conf.paths.branch_images

    @property
    def source_base(self):
        return self.state['source_base']

    @property
    def source_core(self):
        return self.state['source_core']

    @property
    def source_file(self):
        return self.source_base + '.svg'

    @property
    def rst_file(self):
        return self.source_base + '.rst'


class ImageOutputSpecification(giza.libgiza.config.RecursiveConfigurationBase):
    @property
    def type(self):
        if 'type' in self.state:
            return self.state['type']
        else:
            return 'web'

    @type.setter
    def type(self, value):
        possible_types = ('print', 'target', 'web', 'offset')

        if value in possible_types:
            self.state['type'] = value
        else:
            raise TypeError("{0} is not in {1}".format(value, possible_types))

    @property
    def tag(self):
        if 'tag' in self.state:
            return self.state['tag']
        else:
            return None

    @tag.setter
    def tag(self, value):
        if isinstance(value, basestring):
            self.state['tag'] = value
        else:
            raise TypeError('{0}: {1}'.format(type(value), value))

    @property
    def target(self):
        if 'target' in self.state:
            return self.state['target']
        else:
            return None

    @target.setter
    def target(self, value):
        if isinstance(value, basestring):
            self.state['target'] = value
        else:
            raise TypeError('{0}: {1}'.format(type(value), value))

    @property
    def dpi(self):
        if 'dpi' in self.state:
            return self.state['dpi']
        else:
            return 96

    @dpi.setter
    def dpi(self, value):
        if isinstance(value, numbers.Number):
            self.state['dpi'] = value
        else:
            raise TypeError('{0}: {1}'.format(type(value), value))

    @property
    def width(self):
        if 'width' in self.state:
            return self.state['width']
        else:
            return None

    @width.setter
    def width(self, value):
        if isinstance(value, numbers.Number):
            self.state['width'] = value
        else:
            raise TypeError('{0}: {1}'.format(type(value), value))

    @property
    def scale(self):
        if 'scale' in self.state:
            return self.state['scale']
        else:
            return None

    @scale.setter
    def scale(self, value):
        if isinstance(value, numbers.Number) and 100 >= value >= 0:
            self.state['scale'] = value
        else:
            raise TypeError('{0}: {1}'.format(type(value), value))

    @property
    def build_type(self):
        if self.type == 'offset':
            return 'eps'
        else:
            return 'png'

    @property
    def output(self):
        if 'output' not in self.state:
            fn = [self.conf.source_base]

            if self.tag is not None:
                fn.extend(['-', self.tag])

            fn.extend(['.', self.build_type])

            self.state['output'] = ''.join(fn)

        return self.state['output']



# Copyright 2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Post-processing for Sphinx's ``singlehtml`` builder. Modifies links and
post-processes sites for projects that use ``contents.txt`` rather than
``index.txt`` as the root document.
"""

import os.path
import logging
import re

import giza.libgiza.task

from giza.tools.files import expand_tree, copy_if_needed, safe_create_directory
from giza.tools.transformation import decode_lines_from_file, encode_lines_to_file

logger = logging.getLogger('giza.content.post.singlehtml')


def get_single_html_dir(conf):
    return os.path.join(conf.paths.projectroot, conf.paths.public_site_output, 'single')


def manual_single_html(input_file, output_file):
    # don't rebuild this if its not needed.
    if giza.libgiza.task.check_dependency(output_file, input_file) is False:
        logging.info('singlehtml not changed, not reprocessing.')
        return False
    else:
        text_lines = decode_lines_from_file(input_file)

        regexes = [
            (re.compile('href="contents.html'), 'href="index.html'),
            (re.compile('name="robots" content="index"'), 'name="robots" content="noindex"'),
            (re.compile('href="genindex.html'), 'href="../genindex/')
        ]

        for regex, subst in regexes:
            text_lines = [regex.sub(subst, text) for text in text_lines]

        encode_lines_to_file(output_file, text_lines)

        logger.info('processed singlehtml file.')


def finalize_single_html(single_html_dir, artifact_dir, conf):
    for fn in [os.path.join(artifact_dir, f) for f in ('contents.html', 'index.html')]:
        src_fn = os.path.join(conf.paths.projectroot, conf.paths.branch_output, fn)

        if os.path.exists(src_fn):
            manual_single_html(input_file=src_fn,
                               output_file=os.path.join(single_html_dir, 'index.html'))

            copy_if_needed(source_file=os.path.join(artifact_dir, 'objects.inv'),
                           target_file=os.path.join(single_html_dir, 'objects.inv'))


def finalize_single_html_tasks(builder, conf):
    single_html_dir = get_single_html_dir(conf)

    # create directory when registering tasks.
    safe_create_directory(single_html_dir)
    safe_create_directory(os.path.join(single_html_dir, '_static'))

    if 'edition' in conf.project and conf.project.edition != conf.project.name:
        artifact_dir = os.path.join(conf.paths.projectroot,
                                    conf.paths.branch_output,
                                    '-'.join((builder, conf.project.edition)))
    else:
        artifact_dir = os.path.join(conf.paths.projectroot, conf.paths.branch_output, builder)

    tasks = [giza.libgiza.task.Task(job=finalize_single_html,
                                    args=(single_html_dir, artifact_dir, conf),
                                    target=True,
                                    dependency=None,
                                    description="migrating singlehtml")]

    for fn in expand_tree(os.path.join(artifact_dir, '_static'), None):
        target_fn = os.path.join(single_html_dir, '_static', os.path.basename(fn))

        tasks.append(giza.libgiza.task.Task(job=copy_if_needed,
                                            args=(fn, target_fn),
                                            target=target_fn,
                                            dependency=fn,
                                            description="moving static files to the singlehtml build"))

    return tasks



# Copyright 2015 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import giza.libgiza.task
import giza.tools.files
import giza.tools.transformation

logger = logging.getLogger('giza.content.system')


def log_migration(source, target):
    logger.info('migration: {0} --> {1}'.format(source, target))


def migration_tasks(conf):
    tasks = []

    migration_spec_files = conf.system.files.get_configs('migration')

    if 'migrations' not in conf.system.files.data:
        return []

    for migration in conf.system.files.data.migrations:
        copy_job = giza.libgiza.task.Task(job=giza.tools.files.copy_if_needed,
                                          args=(migration.source, migration.target),
                                          target=migration.target,
                                          dependency=migration.source)

        copy_job.finalizers = giza.libgiza.task.Task(job=log_migration,
                                                     args=(migration.source, migration.target))

        if migration.truncate is not None:
            # this only needs to run if the parent task runs
            copy_job.finalizers = giza.libgiza.task.Task(job=giza.tools.transformation.truncate_file,
                                                         args=(migration.target,
                                                               migration.truncate.start_after,
                                                               migration.truncate.end_before),
                                                         target=migration.target,
                                                         dependency=migration.source)

        if migration.transform is not None:
            # causes needs_rebuild() to be always true, this must always run
            copy_job.dependency = None
            copy_job.job = giza.tools.files.copy_always

            regexes = [(t.regex, t.replace) for t in migration.transform]
            transforms = giza.tools.transformation.process_page_task(fn=migration.target,
                                                                     output_fn=migration.target,
                                                                     regex=regexes,
                                                                     builder='migration',
                                                                     copy='ifNeeded')
            copy_job.finalizers = transforms

        if migration.append is not None:
            # causes needs_rebuild() to be always true, this must always run
            copy_job.dependency = None
            copy_job.job = giza.tools.files.copy_always
            copy_job.finalizers = giza.libgiza.task.Task(job=giza.tools.transformation.append_to_file,
                                                         args=(migration.target, migration.append),
                                                         target=migration.target,
                                                         dependency=migration_spec_files)

        tasks.append(copy_job)

    logger.debug('created {0} file migration tasks'.format(len(tasks)))

    return tasks


def migration_clean(conf):
    return [giza.libgiza.task.Task(job=giza.tools.files.verbose_remove,
                                   args=[migration.target],
                                   target=migration.target,
                                   dependency=None)
            for migration in conf.system.files.data.migrations]



# Copyright 2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

from unittest import TestCase

from giza.libgiza.inheritance import (DataContentBase, DataCache,
                                      InheritableContentError, InheritableContentBase)

from giza.config.main import Configuration
from giza.config.runtime import RuntimeStateConfig

# these helper functions are useful in reusing these tests for the giza
# application tests.


def get_inheritance_data_files():
    return [os.path.join(get_test_file_path(), fn)
            for fn in ('example-add-one.yaml', 'example-add-two.yaml',
                       'example-add-three.yaml')]


def get_test_file_path():
    return os.path.abspath(os.path.join(os.path.dirname(__file__), 'data-inheritance'))


class TestDataCache(TestCase):
    def setUp(self):
        self.c = Configuration()
        self.c.runstate = RuntimeStateConfig()
        self.c.paths = {'includes': get_test_file_path()}

        self.DataContentBase = DataContentBase
        self.DataCache = DataCache

        self.create_data()

    def create_data(self):
        self.files = get_inheritance_data_files()
        self.data = self.DataCache([], self.c)

    def test_content_class_default(self):
        self.assertEqual(self.data.cache, {})
        self.assertIs(self.data.content_class, self.DataContentBase)

    def test_cache_not_setable(self):
        self.assertEqual(self.data.cache, {})
        self.data.cache = []
        self.assertEqual(self.data.cache, {})

    def test_cache_property_reference(self):
        self.assertIs(self.data.cache, self.data._cache)

    def test_cache_cleaner(self):
        self.assertEqual(self.data.cache, {})

        self.data._clear_cache('foo')
        self.data._clear_cache('bar')

        self.assertIn('foo', self.data.cache)
        self.assertIn('bar', self.data.cache)

        self.assertEqual(self.data.cache['foo'], [])
        self.assertEqual(self.data.cache['bar'], [])

    def test_membership(self):
        self.assertEqual(self.data.cache, {})

        self.data._clear_cache('foo')
        self.data._clear_cache('bar')

        self.assertIn('foo', self.data.cache)
        self.assertIn('bar', self.data.cache)

        self.assertIn('foo', self.data)
        self.assertIn('bar', self.data)

    def test_ingest(self):
        self.assertEqual(self.data.cache, {})
        self.data.ingest(self.files)
        self.assertNotEqual(self.data.cache, {})

        for fn in self.files:
            self.assertIn(fn, self.data)

    def test_ingest_ignore_duplicates(self):
        self.assertEqual(self.data.cache, {})
        self.data.ingest(self.files)
        self.data.ingest(self.files)
        self.assertNotEqual(self.data.cache, {})

        self.assertEqual(len(self.data.cache), len(self.files))

    def test_add_file(self):
        self.assertEqual(self.data.cache, {})

        for fn in self.files:
            self.data.add_file(fn)
            self.assertIn(fn, self.data.cache)
            self.assertIn(fn, self.data)
            self.assertIsInstance(self.data.cache[fn], self.data.content_class)

    def test_fetch(self):
        self.assertEqual(self.data.cache, {})
        self.data.ingest(self.files)

        for fn in self.files:
            self.assertIn(fn, self.data)
            for idx in range(len(self.data.cache), 1):
                content = self.data.fetch(fn, idx)
                self.assertIsInstance(content, dict)

        self.assertEqual(len(self.data.cache), len(self.files))

    def test_fetch_without_adding_file(self):
        self.assertEqual(self.data.cache, {})

        for fn in self.files:
            if not os.path.isfile(fn):
                with self.assertRaises(InheritableContentError):
                    self.data.fetch(fn, 1)
                self.assertNotIn(fn, self.data)


class TestDataContentBase(TestCase):
    def setUp(self):
        self.c = Configuration()
        self.c.runstate = RuntimeStateConfig()
        self.c.paths = {'includes': 'data-inheritance',
                        'projectroot': os.path.abspath(os.path.dirname(__file__))}

        self.DataContentBase = DataContentBase
        self.DataCache = DataCache
        self.InheritableContentBase = InheritableContentBase

        self.dummy_doc = {'ref': 'dummy-doc',
                          'pre': 'pre text'}

        self.create_data()

    def create_data(self):
        self.content_fn = get_inheritance_data_files()[0]

        self.data = self.DataCache([self.content_fn], self.c)
        self.content = self.data.cache[self.content_fn]

    def test_content_created(self):
        for fn, example in self.data.file_iter():
            self.assertIsInstance(example, self.DataContentBase)

    def test_content_item_created(self):
        for fn, block in self.data.content_iter():
            self.assertIsInstance(block, self.InheritableContentBase)

    def test_content_is_correct_type(self):
        self.assertIsInstance(self.content, self.DataContentBase)

    def test_content_state_reference(self):
        self.assertIs(self.content.content, self.content.state['content'])

    def test_reference_to_cache(self):
        self.assertIs(self.content.data, self.data)

    def test_fetching_content(self):
        for idx in self.content.content:
            compared = self.data.fetch(self.content_fn, idx)
            for key in self.content.content[idx].state:
                self.assertEqual(self.content.content[idx].state[key], getattr(compared, key))

    def test_resolve_checker(self):
        if isinstance(self.content, dict):
            with self.assertRaises(AttributeError):
                self.content.is_resolved()
        else:
            self.assertIsInstance(self.content.is_resolved(), bool)

    def test_add_always_returns_value(self):
        cache = self.DataCache([], self.c)
        collection = cache.content_class('', cache, self.c)
        self.assertIsInstance(collection, DataContentBase)

        content = collection.add(self.dummy_doc)

        self.assertIsInstance(content, InheritableContentBase)
        print(collection.content)


class TestInheritedContentResolution(TestCase):
    def setUp(self):
        self.c = Configuration()
        self.c.runstate = RuntimeStateConfig()
        self.c.paths = {'includes': os.path.join(os.path.abspath(os.path.dirname(__file__)),
                                                 'data-inheritance')}
        self.DataCache = DataCache

        self.create_data()

    def create_data(self):
        self.data = self.DataCache(get_inheritance_data_files(), self.c)
        self.len_source_docs = 5
        self.num_docs = 3

    def test_gross_correctness_of_ingestion(self):
        self.assertEqual(len(self.data.cache), self.num_docs)

    def test_everything_resolved(self):
        for fn, data in self.data.cache.items():
            self.assertIsInstance(data, self.data.content_class)

            self.assertNotEqual(len(data.content), 0)

            for doc in data.content.values():
                if 'source' in doc:
                    self.assertEqual(len(doc.state.keys()), self.len_source_docs)

            data.resolve()

            for doc in data.content.values():
                if 'source' in doc:
                    self.assertTrue(doc.source.resolved)


class TestBaseTemplateRendering(TestCase):
    def setUp(self):
        self.c = Configuration()
        self.c.runstate = RuntimeStateConfig()
        self.InheritableContentBase = InheritableContentBase

        self.create_data()

    def create_data(self):
        self.data = self.InheritableContentBase({}, self.c)
        self.data.replacement = {'state': 'foo'}

    def test_replacement(self):
        self.data.pre = 'this is a {{state}} test'
        self.assertTrue('{{' in self.data.pre)
        self.data.render()
        self.assertFalse('{{' in self.data.pre)
        self.assertTrue('foo' in self.data.pre)



# Copyright 2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Post processing of the output of the Sphinx manpage output. Used to get proper
fully qualified links and fix flaws in the output.
"""

import re
import os
import logging

import giza.libgiza.task

from giza.tools.files import expand_tree

logger = logging.getLogger('giza.content.manpages')

# Manpage Processing

# this is a precursor to giza.tools.transformation and should be re-implemented.


def manpage_url(regex_obj, input_file):
    with open(input_file, 'r') as f:
        manpage = f.read()

    if isinstance(regex_obj, list):
        for regex, subst in regex_obj:
            manpage = regex.sub(subst, manpage)
    else:
        manpage = regex_obj[0].sub(regex_obj[1], manpage)

    with open(input_file, 'w') as f:
        f.write(manpage)

    logger.info("fixed urls in {0}".format(input_file))


def manpage_url_tasks(builder, conf):
    project_source = os.path.join(conf.paths.projectroot,
                                  conf.paths.source)

    top_level_items = set()
    for fs_obj in os.listdir(project_source):
        if fs_obj.startswith('.static') or fs_obj == 'index.txt':
            continue
        if os.path.isdir(os.path.join(project_source, fs_obj)):
            top_level_items.add(fs_obj)
        if fs_obj.endswith('.txt'):
            top_level_items.add(fs_obj[:-4])

    top_level_items = '/' + r'[^\s]*|/'.join(top_level_items) + r'[^\s]*'

    re_string = r'(\\fB({0})\\fP)'.format(top_level_items).replace(r'-', r'\-')
    subst = conf.project.url + '/' + conf.project.tag + r'\2'

    regex_obj = (re.compile(re_string), subst)

    tasks = []
    for manpage in expand_tree(os.path.join(conf.paths.projectroot,
                                            conf.paths.output,
                                            conf.git.branches.current,
                                            builder), ['1', '5']):

        description = 'processing urls in manpage file: {0}'.format(manpage)
        tasks.append(giza.libgiza.task.Task(job=manpage_url,
                                            args=(regex_obj, manpage),
                                            target=manpage,
                                            dependency=None,
                                            description=description))

    return tasks



# Copyright 2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
:mod:`giza.operations.make` provides an alternate method for running build
operations, by specifying operations as a list of strings that resemble ``make``
targets. These targets are compatible with the targets created in the default
``sphinx-quickstart`` ``makefile``, and also supports a number of additional
options for specifying "editions" and human-language variants of build
operations, as well as deployment ("push") and packaging operations.

Many projects retain a default makefile that passes all make operations through
to this mode of giza operation, with the following target:

.. code-block:: makefile

   %:
       giza make %@
"""

import logging
import copy
import argh
import itertools

from giza.libgiza.app import BuildApp
from giza.libgiza.task import Task
from giza.config.helper import fetch_config
from giza.config.sphinx_config import available_sphinx_builders
from giza.operations.deploy import deploy_tasks
from giza.operations.sphinx_cmds import sphinx_publication
from giza.operations.build_env import env_package_worker

logger = logging.getLogger('giza.operations.make')


def derive_command(name, conf):
    """
    :param str name: The name of a build operation. Currently supports
        ``sphinx``, ``env``, ``deploy``. Unknown operations are ignored.

    :param Configuration conf: The top level configuration object.

    Given a ``conf`` object and the name of the operation, logs a message with a
    complete specific giza command that describes an equivalent specific command.
    """

    if name in ('sphinx', 'env'):
        cmd = ["giza"]
        if conf.runstate.fast is True:
            cmd.append('--fast')

        cmd.append(name)

        if conf.runstate.serial_sphinx is True:
            cmd.append('--serial_sphinx')

        if len(conf.runstate.builder) > 0:
            cmd.append('--builder')
            cmd.append(' '.join(conf.runstate.builder))

        if (len(conf.runstate.editions_to_build) > 0 and
                conf.runstate.editions_to_build[0] is not None):

            cmd.append('--edition')
            cmd.append(' '.join(conf.runstate.editions_to_build))

        if (len(conf.runstate.languages_to_build) > 0 and
                conf.runstate.languages_to_build[0] is not None):

            cmd.append('--language')
            cmd.append(' '.join(conf.runstate.languages_to_build))

        if name == 'sphinx':
            logger.info('running sphinx build operation, equivalent to: ' + ' '.join(cmd))
        elif name == 'env':
            logger.info('running env cache operation, equivalent to: ' + ' '.join(cmd))
    elif name == 'deploy':
        cmd = ["giza deploy"]
        cmd.append('--target')
        cmd.append(' '.join(conf.runstate.push_targets))

        logger.info('running deploy operation, equivalent to: ' + ' '.join(cmd))


def add_sphinx_build_options(action_spec, action, options, conf):
    """
    :param dict action_spec: A reference to a dictionary that defines a

    Modifies the ``action_spec`` document to define a build operation with the
    appropriate options, based on the operation (``action``) and the
    ``options``.

    If the options do not specify one or more editions in a project that uses
    editions, will add all configured editions to the build operation.
    """

    sphinx_builders = available_sphinx_builders()

    if action in sphinx_builders:
        action_spec['builders'].add(action)

    for build_option in options:
        if build_option in sphinx_builders:
            action_spec['builders'].add(build_option)
        elif build_option in conf.project.edition_list:
            action_spec['editions'].add(build_option)
        elif build_option in conf.system.files.data.integration:
            action_spec['languages'].add(build_option)
        elif build_option == 'fast' and action != 'publish' and 'publish' not in options:
            conf.runstate.fast = True

    if 'editions' in conf.project and len(action_spec['editions']) == 0:
        action_spec['editions'].update(conf.project.edition_list)


@argh.arg('make_target', nargs="*")
@argh.arg('--serial_sphinx', action='store_true')
@argh.named('make')
@argh.expects_obj
def main(args):
    """
    Provides a way to specify make-like targets to invoke giza
    operations. Targets take a <action>-<option<-option>> form.
    """
    targets = [(t[0], t[1:]) for t in [t.split("-") for t in args.make_target]]

    conf = fetch_config(args)
    run_make_operations(targets, conf)


def run_make_operations(targets, conf):
    """
    :param list targets: A list of tuples in the form of ``(<action>, [option,
         option])`` that define build targets.

    :param Configuration conf: The top level configuration object.

    Parses the ``targets`` list and runs tasks defined, including all specified
    sphinx targets, all ``push`` deployment targets, and will create the ``env``
    packages. Noteworthy behavior:

    - The order of options *except* for the action in the first option is not
      important.

    - If you run ``push`` target with the ``deploy`` option
      (i.e. ``push-deploy`` or ``push-<edition>-deploy``), ``giza`` will *not*
      run the ``publish`` Sphinx build.

    - This interface assumes that all deployment targets (defined in each
      project begin with ``push-`` or ``stage-``.) If you have a project with
      different deployment targets, you will need to call ``giza deploy``
      directly.

    - The ``env`` cache targets take the same options as the Sphinx builders and
      package the environment for only those builders. If you specify ``env``
      after a Sphinx target, ``giza`` will build the cache for only that
      package.
    """

    sphinx_opts = {"languages": set(),
                   "editions": set(),
                   "builders": set()}
    push_opts = {"targets": set(),
                 "type": None}
    packaging_opts = {}

    sphinx_builders = available_sphinx_builders()

    if 'push' in conf.system.files.data:
        deploy_configs = dict((item['target'], item) for item in conf.system.files.data.push)
    else:
        deploy_configs = []

    tasks = []
    for action, options in targets:
        if action in sphinx_builders:
            tasks.append(sphinx_opts)

            add_sphinx_build_options(sphinx_opts, action, options, conf)
        elif action in ('stage', 'push'):
            tasks.append(push_opts)
            push_opts['type'] = action

            if 'deploy' not in options:
                sphinx_opts['builders'].add('publish')
                tasks.append(sphinx_opts)
                add_sphinx_build_options(sphinx_opts, action, options, conf)
                conf.runstate.fast = False

            if action in deploy_configs:
                push_opts['targets'].add(action)

            for build_option in options:
                deploy_target_name = '-'.join((action, build_option))

                if build_option in deploy_configs:
                    push_opts['targets'].add(build_option)
                elif deploy_target_name in deploy_configs:
                    push_opts['targets'].add(deploy_target_name)
        elif action.startswith('env'):
            if len(packaging_opts) > 0:
                packaging_opts = copy.copy(sphinx_opts)

            tasks.append(packaging_opts)
            add_sphinx_build_options(packaging_opts, False, options, conf)
        else:
            logger.error('target: {0} not defined in the make interface'.format(action))

    app = BuildApp.new(pool_type=conf.runstate.runner,
                       force=conf.runstate.force,
                       pool_size=conf.runstate.pool_size)

    if sphinx_opts in tasks:
        conf.runstate.languages_to_build = list(sphinx_opts['languages'])
        conf.runstate.editions_to_build = list(sphinx_opts['editions'])
        conf.runstate.builder = list(sphinx_opts['builders'])

        if 'publish' in conf.runstate.builder:
            conf.runstate.fast = False

        derive_command('sphinx', conf)

        sphinx_publication(conf, app)

    if push_opts in tasks:
        if len(push_opts['targets']) == 0:
            for lang, edition in itertools.product(conf.runstate.languages_to_build,
                                                   conf.runstate.editions_to_build):
                push_target_name = [push_opts['type']]
                for opt in (edition, lang):
                    if opt is not None:
                        push_target_name.append(opt)
                push_target_name = '-'.join(push_target_name)
                push_opts['targets'].add(push_target_name)

        conf.runstate.push_targets = list(push_opts['targets'])
        deploy_tasks(conf, app)
        derive_command('deploy', conf)

    if packaging_opts in tasks:
        derive_command('env', conf)

        app.add(Task(job=env_package_worker,
                     args=(conf.runstate, conf),
                     target=True,
                     dependency=None))

    if len(app.queue) >= 1:
        app.run()



# Copyright 2014 MongoDB, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys
import logging

from giza.libgiza.config import ConfigurationBase
from giza.config.runtime import RuntimeStateConfigurationBase

logger = logging.getLogger('giza.config.github')

if sys.version_info >= (3, 0):
    basestring = str


def fetch_config(args):
    c = GithubConfig()
    c.ingest(args.conf_path)
    c.runstate = args

    return c


class GithubRuntimeConfig(RuntimeStateConfigurationBase):

    @property
    def conf_path(self):
        if 'conf_path' not in self.state:
            self.conf_path = None

        return self.state['conf_path']

    @conf_path.setter
    def conf_path(self, value):
        if value is not None and os.path.exists(value):
            self.state['conf_path'] = value
        else:
            try:
                self._discover_conf_file('.github.yaml')
            except OSError:
                self._discover_conf_file('github.yaml')
            except OSError:
                logger.error('could not find mdbpr github config file.')
                raise OSError


class GithubConfig(ConfigurationBase):

    @property
    def runstate(self):
        return self.state['runstate']

    @runstate.setter
    def runstate(self, value):
        if isinstance(value, GithubRuntimeConfig):
            value.conf = self
            self.state['runstate'] = value
        else:
            msg = "invalid runtime state"
            logger.critical(msg)
            raise TypeError(msg)

    @property
    def site(self):
        return self.state['site']

    @site.setter
    def site(self, value):
        if isinstance(value, GithubSiteConfig):
            self.state['site'] = value
        else:
            self.state['site'] = GithubSiteConfig(value)

    @property
    def repos(self):
        return self.state['repos']

    @repos.setter
    def repos(self, value):
        if isinstance(value, list):
            self.state['repos'] = []
            for repo in value:
                self.state['repos'].append(GithubRepoConfig(repo))
        elif isinstance(value, dict):
            if 'name' in value and 'user' in value:
                self.state['repos'].append(GithubRepoConfig(repo))
        else:
            raise TypeError('{0} is not a valid repo sepc'.format(value))

    @property
    def organizations(self):
        return self.state['organizations']

    @organizations.setter
    def organizations(self, value):
        if isinstance(value, list):
            self.state['organizations'] = value
        elif isinstance(value, basestring):
            self.state['organizations'] = [value]
        else:
            raise TypeError('{0} is not a valid organization or organization list'.format(value))

    @property
    def reporting(self):
        return self.state['reporting']

    @reporting.setter
    def reporting(self, value):
        if isinstance(value, GithubReportingConfig):
            self.state['reporting'] = value
        else:
            self.state['reporting'] = GithubReportingConfig(value)


class GithubRepoConfig(ConfigurationBase):
    _option_registry = ['user', 'name']


class GithubSiteConfig(ConfigurationBase):
    _option_registry = ['corp']

    @property
    def credentials(self):
        return self.state['credentials']

    @credentials.setter
    def credentials(self, value):
        value = os.path.expanduser(value)
        self.state['credentials'] = value


class GithubReportingConfig(ConfigurationBase):
    _option_registry = ['format']



"""
Provides the basis for a more strictly typed dictionary that ensures all keys
and values will be of a specific type, and will throw exceptions if keys,
values, or pairs of keys and values are of the wrong type or fail validation.
"""

import abc
import future.utils
import logging

import giza.libgiza.error

logger = logging.getLogger('giza.libgiza.typed_dict')


class TypedDict(future.utils.with_metaclass(abc.ABCMeta, dict)):
    """
    An abstract base class definition that ensures that keys and values are of
    the correct type. Requires users implement ``check_key()``,
    ``check_value()`` and ``check_pair()`` methods that allow the these objects
    to validate input.
    """

    def __init__(self, key_type, value_type):
        errors = giza.libgiza.error.ErrorCollector()
        if isinstance(key_type, type):
            self.key_type = key_type
        else:
            errors.add(giza.libgiza.error.Error(
                message="key_type ({0}) is not a type value".format(key_type)))

        if isinstance(value_type, type):
            self.value_type = value_type
        else:
            errors.add(giza.libgiza.error.Error(
                "value_type ({0}) is not a type value".format(value_type)))

        if errors.has_errors() and errors.fatal:
            logger.debug(errors.render_output())
            raise TypeError(errors.dict())

    def __setitem__(self, key, value):
        type_errors = giza.libgiza.error.ErrorCollector()
        value_errors = giza.libgiza.error.ErrorCollector()

        if isinstance(key, self.key_type):
            value_errors.add(self.check_key(key))
        else:
            try:
                key = self.key_type(key)
            except Exception as e:
                type_errors.add(giza.libgiza.error.ErrorCollector(
                    message=("key {0} ({1}) is not of type {2} (had error "
                             "{3}:{4})").format(key, type(key), self.key_type, type(e), e)))

        if isinstance(value, self.value_type):
            value_errors.add(self.check_value(value))
        else:
            try:
                value = self.value_type(value)
            except Exception as e:
                type_errors.add(giza.libgiza.error.ErrorCollector(
                    message=("value for key {0} is not of type {1} (is {2}). (had error "
                             "{3}:{4})").format(key, self.value_type, type(value), type(e), e)))

        # if checks for pair errors depend on type/values being correct they
        # may except in unpredictable ways
        try:
            value_errors.add(self.check_pair(key, value))
        except Exception as e:
            value_errors.add(giza.libgiza.error.Error(
                message=("encountered {0} error when validating "
                         "pair for key {1}").format(type(e), key)))

        if type_errors.has_errors() and type_errors.fatal:
            logger.debug(type_errors.render_output())
            if value_errors.has_errors():
                logger.debug(value_errors.render_output())
            raise TypeError(type_errors.dict())

        if value_errors.has_errors() and value_errors.fatal:
            logger.debug(value_errors.render_output())
            raise ValueError(value_errors.dict())

        dict.__setitem__(self, key, value)

    def ingest(self, args):
        if args is None or len(args) == 0:
            return
        elif isinstance(args, tuple):
            dict.__init__(self, *args)
        else:
            dict.__init__(self, args)

    @abc.abstractmethod
    def check_key(self, key):
        return giza.libgiza.error.ErrorCollector()

    @abc.abstractmethod
    def check_value(self, value):
        return giza.libgiza.error.ErrorCollector()

    @abc.abstractmethod
    def check_pair(self, key, value):
        return giza.libgiza.error.ErrorCollector()
